from __future__ import annotations

from inspect import isclass
from pathlib import Path
from typing import Optional, Dict, List, Callable, Union, Type

from pyqtgraph.Qt import QtCore
from pyqtgraph.parametertree import Parameter
from pyqtgraph.parametertree.parameterTypes import ListParameter
from utilitys import ParamEditor, NestedProcWrapper, fns
from utilitys.fns import setParamTooltips
from utilitys.params.pgregistered import ProcGroupParameter
from utilitys import NestedProcess

from s3a.constants import MENU_OPTS_DIR

Signal = QtCore.Signal

class AlgCtorCollection(ParamEditor):
  # sigProcessorCreated = Signal(object) # Signal(AlgCollectionEditor)
  def __init__(self, procWrapType: Type[NestedProcWrapper], parent=None):
    super().__init__(parent, saveDir='', fileType='')
    self.processorCtors : List[Callable[[], NestedProcess]] = []
    self.spawnedCollections : List[AlgParamEditor] = []
    self.procWrapType = procWrapType

  def createProcessorForClass(self, clsObj, editorName='Processor') -> AlgParamEditor:
    if not isclass(clsObj):
      clsObj = type(clsObj)
    clsName = clsObj.__name__
    formattedClsName = fns.pascalCaseToTitle(clsName)
    editorDir = MENU_OPTS_DIR/formattedClsName.lower()
    newEditor = AlgParamEditor(editorDir, self.processorCtors, self.procWrapType, name=editorName)
    self.spawnedCollections.append(newEditor)
    # Wrap in property so changes propagate to the calling class
    lims = newEditor.algOpts.opts['limits']
    defaultKey = next(iter(lims))
    defaultAlg = lims[defaultKey]
    newEditor.algOpts.setDefault(defaultAlg)
    newEditor.switchActiveProcessor(proc=defaultAlg)
    # self.sigProcessorCreated.emit(newEditor)
    return newEditor

  def addProcessCtor(self, procCtor: Callable[[], NestedProcess]):
    self.processorCtors.append(procCtor)
    for algCollection in self.spawnedCollections:
      algCollection.addProcessor(procCtor())

  def addProcessFunction(self, func: Callable, procType: Type[NestedProcess], name:str=None, **kwargs):
    def ctor():
      return procType.fromFunction(func, name=name, **kwargs)
    self.addProcessCtor(ctor)

class AlgParamEditor(ParamEditor):
  def __init__(self, saveDir, procCtors: List[Callable[[], NestedProcess]],
               procWrapType: Type[NestedProcWrapper], name=None, parent=None):
    algOptDict = {
      'name': 'Algorithm', 'type':  'list', 'values': [], 'value': 'N/A'
    }
    self.treeAlgOpts: Parameter = Parameter(name='Algorithm Selection', type='group', children=[algOptDict])
    self.algOpts: ListParameter = self.treeAlgOpts.children()[0]
    self.nameToProcMapping: Dict[str, NestedProcWrapper] = {}
    super().__init__(parent, saveDir=saveDir, fileType='alg', name=name,
                     topTreeChild=self.algOpts)
    self.algOpts.sigValueChanged.connect(lambda param, proc: self.switchActiveProcessor(proc))
    self.expandAllBtn.hide()
    self.collapseAllBtn.hide()

    self.saveDir.mkdir(parents=True, exist_ok=True)

    self.curProcessor: Optional[NestedProcWrapper] = None
    self.procWrapType = procWrapType

    wrapped : Optional[NestedProcWrapper] = None
    for processorCtor in procCtors:
      # Retrieve proc so default can be set after
      wrapped = self.addProcessor(processorCtor())
    self.algOpts.setDefault(wrapped)
    self.switchActiveProcessor(proc=wrapped)
    # self.saveParamValues('Default', allowOverwriteDefault=True)

  def addProcessor(self, newProc: NestedProcess):
    processor = self.procWrapType(newProc, parentParam=self.params)
    self.tree.addParameters(self.params.child(processor.algName))
    setParamTooltips(self.tree)

    self.nameToProcMapping.update({processor.algName: processor})
    self.algOpts.setLimits(self.nameToProcMapping.copy())
    return processor

  def saveParamValues(self, saveName: str=None, paramState: dict=None, **kwargs):
    """
    The algorithm editor also needs to store information about the selected algorithm, so lump
    this in with the other parameter information before calling default save.
    """
    if paramState is None:
      paramDict = fns.paramDictWithOpts(self.params, addList=['enabled'], addTo=[ProcGroupParameter],
                                         removeList=['value'])
      paramState = {'Selected Algorithm': self.algOpts.value().algName,
                    'Parameters': paramDict}
    return super().saveParamValues(saveName, paramState, **kwargs)

  def loadParamValues(self, stateName: Union[str, Path],
                      stateDict: dict=None, **kwargs):
    stateDict = self._parseStateDict(stateName, stateDict)
    selectedOpt = stateDict.get('Selected Algorithm', None)
    # Get the impl associated with this option name
    isLegitSelection = selectedOpt in self.algOpts.opts['limits']
    if not isLegitSelection:
      selectedImpl = self.algOpts.value()
      raise ValueError(f'Selection {selectedOpt} does'
                                f' not match the list of available algorithms. Defaulting to {selectedImpl}')
    else:
      selectedImpl = self.algOpts.opts['limits'][selectedOpt]
    self.algOpts.setValue(selectedImpl)
    super().loadParamValues(stateName, stateDict['Parameters'])

  def switchActiveProcessor(self, proc: Union[str, NestedProcWrapper]):
    """
    Changes which processor is active. if ImgProcWrapper, uses that as the processor.
    If str, looks for that name in current processors and uses that
    """
    if isinstance(proc, str):
      proc = self.nameToProcMapping[proc]
    # Hide all except current selection
    # TODO: Find out why hide() isn't working. Documentation indicates it should
    # Instead, use the parentChanged utility as a hacky workaround
    selectedParam = self.params.child(proc.algName)
    for ii, child in enumerate(self.params.children()):
      shouldHide = child is not selectedParam
      # Offset by 1 to account for self.algOpts
      self.tree.setRowHidden(1 + ii, QtCore.QModelIndex(), shouldHide)
    # selectedParam.show()
    self.curProcessor = proc


from __future__ import annotations

import pydoc
import shutil
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import Optional, Set, List, Dict, Sequence, Union, Tuple
from warnings import warn

import numpy as np
import pandas as pd
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets, QtCore
from skimage import io
from utilitys import CompositionMixin, AtomicProcess
from utilitys import fns
from utilitys.fns import warnLater
from utilitys.params import *

from s3a import PRJ_SINGLETON, PRJ_CONSTS as CNST, ComponentIO, models, REQD_TBL_FIELDS
from s3a.generalutils import hierarchicalUpdate, cvImsave_rgb
from s3a.graphicsutils import DropList, ThumbnailViewer
from s3a.structures import FilePath, NChanImg
from ..constants import APP_STATE_DIR, PROJ_FILE_TYPE, PROJ_BASE_TEMPLATE


class FilePlugin(CompositionMixin, ParamEditorPlugin):
  name = '&File'
  win: models.s3abase.S3A
  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()

  def __init__(self, startupName: FilePath=None, startupCfg: dict=None):
    super().__init__()
    self.projData = self.exposes(ProjectData(startupName, startupCfg))
    self.autosaveTimer = QtCore.QTimer()

    self.registerFunc(self.save, btnOpts=CNST.TOOL_PROJ_SAVE)
    self.registerFunc(self.showProjImgs_gui, btnOpts=CNST.TOOL_PROJ_OPEN_IMG)
    self.menu.addSeparator()

    self.registerFunc(self.create_gui, btnOpts=CNST.TOOL_PROJ_CREATE)
    self.registerFunc(self.open_gui, btnOpts=CNST.TOOL_PROJ_OPEN)

    self.registerPopoutFuncs([self.updateProjectProperties, self.addImages_gui, self.addAnnotations_gui],
                             ['Update Project Properties', 'Add Images', 'Add Annotations'],
                             btnOpts=CNST.TOOL_PROJ_SETTINGS)

    self.registerFunc(lambda: self.win.setMainImg_gui(), btnOpts=CNST.TOOL_PROJ_ADD_IMG)
    self.registerFunc(lambda: self.win.openAnnotation_gui(), btnOpts=CNST.TOOL_PROJ_ADD_ANN)

    self.registerPopoutFuncs([self.startAutosave, self.stopAutosave], btnOpts=CNST.TOOL_AUTOSAVE)

    self._projImgMgr = ProjectImageManager()
    self._imgThumbnails = self._projImgMgr.thumbnails
    self.nameToFullPathMapping = self._imgThumbnails.nameToFullPathMapping

    def onAdd(imList):
      for im in imList:
        self._imgThumbnails.addThumbnail(im, force=True)
    def onMove(imList):
      for oldName, newName in imList:
        self.nameToFullPathMapping[oldName.name] = newName
    def onDel(imList):
      if len(imList) == 0:
        # or QtWidgets.QMessageBox.question(
        # self.win, 'Confirm Delete', f'Are you sure you want to delete the following images?\n'
        #                             f'{[i.name for i in imList]})') != QtWidgets.QMessageBox.Yes:
        return
      delCurrent = False
      for name in imList:
        if name == self.win.srcImgFname:
          delCurrent = True
          continue
        self._imgThumbnails.removeThumbnail(name.name)
      if delCurrent:
        raise IOError(f'Cannot delete {self.win.srcImgFname.name} since it is currently'
             f' being annotated. Change the image and try again.')
    self.projData.sigImagesAdded.connect(onAdd)
    self.projData.sigImagesMoved.connect(onMove)
    self.projData.sigImagesRemoved.connect(onDel)

    self._projImgMgr.sigImageSelected.connect(lambda imgFname: self.win.setMainImg(imgFname))
    def handleDelete(delImgs):
      for img in delImgs:
        self.projData.removeImage(img)
    self._projImgMgr.sigDeleteRequested.connect(handleDelete)

    def onCfgLoad():
      for im in self._imgThumbnails.nameToFullPathMapping.values():
        if im not in self.projData.images:
          self._imgThumbnails.removeThumbnail(im.name)
      self._updateProjLbl()
      if self.win:
        # Other arguments are consumed by app state editor
        state = self.win.appStateEditor
        if state.loading:
          hierarchicalUpdate(state.startupSettings, self.projData.startup)
        else:
          # Opening a project after s3a is already loaded
          state.loadParamValues(stateDict={}, **self.projData.startup)

    self.projData.sigCfgLoaded.connect(onCfgLoad)

    self.projNameLbl = QtWidgets.QLabel()

    useDefault = startupName is None and startupCfg is None
    self._createDefaultProj(useDefault)

  def _updateProjLbl(self):
    self.projNameLbl.setText(f'Project: {self.projData.cfgFname.name}')

  def attachWinRef(self, win: models.s3abase.S3ABase):
    super().attachWinRef(win)
    win.statBar.addWidget(self.projNameLbl)
    def handleExport(_dir):
      saveImg = win.srcImgFname
      ret = str(self.projData.cfgFname)
      if not saveImg:
        self.projData.startup.pop('image', None)
        return ret
      if saveImg and saveImg.parent == self.projData.imagesDir:
        saveImg = saveImg.name
      self.projData.startup['image'] = str(saveImg)
      self.save()
      return ret
    win.appStateEditor.addImportExportOpts('project', self.open, handleExport, 0)

    def startImg(imgName: str):
      self.projData.addImage(imgName)
      name = Path(imgName).name
      fullName = self.nameToFullPathMapping[name]
      self.win.setMainImg(fullName)
    win.appStateEditor.addImportExportOpts('image', startImg, lambda *args: None, 1)

    doctored = AtomicProcess(self.win.exportCurAnnotation, 'Current Annotation',
                             outFname='')
    self.registerPopoutFuncs([self.projData.exportProj, self.projData.exportAnnotations, doctored],
                         ['Project', 'All Annotations', 'Current Annotation'], btnOpts=CNST.TOOL_PROJ_EXPORT)

    self._projImgMgr.hide()
    self._updateProjLbl()
    win.addTabbedDock(QtCore.Qt.RightDockWidgetArea, self._projImgMgr)

  def _createDefaultProj(self, setAsCur=True):
    defaultName = APP_STATE_DIR / PROJ_FILE_TYPE
    # Delete default prj on startup
    for prj in defaultName.glob(f'*.{PROJ_FILE_TYPE}'):
      prj.unlink()
    parent = self.projData if setAsCur else None
    self.projData.create(name=defaultName, parent=parent)

  def open(self, cfgFname: FilePath=None, cfgDict: dict=None):
    if cfgFname is not None and Path(cfgFname).resolve() != self.projData.cfgFname:
      self.win.setMainImg(None)
      self.projData.loadCfg(cfgFname, cfgDict)

  def open_gui(self):
    fname = fns.popupFilePicker(None, 'Select Project File', f'S3A Project (*.{PROJ_FILE_TYPE})')
    with pg.BusyCursor():
      self.open(fname)

  def save(self):
    self.win.saveCurAnnotation()
    self.projData.saveCfg()

  @fns.dynamicDocstring(ioTypes=['<Unchanged>'] + list(ComponentIO.handledIoTypes))
  def updateProjectProperties(self, tableConfig:FilePath=None, annotationFormat:str=None):
    """
    Updates the specified project properties, for each one that is provided

    :param tableConfig:
      pType: filepicker
    :param annotationFormat:
      helpText: "How to save annotations internally. Note that altering
      this value may alter the speed of saving and loading annotations."
      pType: list
      limits: {ioTypes}
    """
    if tableConfig is not None:
      tableConfig = Path(tableConfig)
      self.projData.tableData.loadCfg(tableConfig)
    if annotationFormat is not None and annotationFormat in ComponentIO.handledIoTypes:
      self.projData.cfg['annotation-format'] = annotationFormat


  @fns.dynamicDocstring(ioTypes=list(ComponentIO.handledIoTypes))
  def startAutosave(self, interval=5, backupFolder='', baseName='autosave', exportType='pkl'):
    """
    Saves the current annotation set evert *interval* minutes

    :param interval:
      helpText: Interval in minutes between saves
      limits: [1, 1e9]
    :param backupFolder:
      helpText: "If provided, annotations are saved here sequentially afte reach *interval* minutes.
      Each output is named `[Parent Folder]/[base name]_[counter].[export type]`, where `counter`
      is the current save file number.'"
      pType: filepicker
      asFolder: True
    :param baseName: What to name the saved annotation file
    :param exportType:
      helpText: File format for backups
      pType: list
      limits: {ioTypes}
    """
    self.autosaveTimer = QtCore.QTimer()
    self.autosaveTimer.start(int(interval * 60 * 1000))
    self.autosaveTimer.timeout.connect(self.win.saveCurAnnotation)
    if len(str(backupFolder)) == 0:
      return
    backupFolder = Path(backupFolder)
    backupFolder.mkdir(exist_ok=True, parents=True)
    lastSavedDf = self.win.exportableDf.copy()
    # Qtimer expects ms, turn mins->s->ms
    # Figure out where to start the counter
    globExpr = lambda: backupFolder.glob(f'{baseName}*.{exportType}')
    existingFiles = list(globExpr())
    if len(existingFiles) == 0:
      counter = 0
    else:
      counter = max(map(lambda fname: int(fname.stem.rsplit('_')[1]), existingFiles)) + 1

    def save_incrementCounter():
      nonlocal counter, lastSavedDf
      baseSaveNamePlusFolder = backupFolder / f'{baseName}_{counter}.{exportType}'
      counter += 1
      curDf = self.win.exportableDf
      if not curDf.equals(lastSavedDf):
        self.win.exportCurAnnotation(baseSaveNamePlusFolder)
        lastSavedDf = curDf.copy()

    self.autosaveTimer.timeout.connect(save_incrementCounter)

  def stopAutosave(self):
    self.autosaveTimer.stop()

  def showProjImgs_gui(self):
    if len(self.projData.images) == 0:
      warn('This project does not have any images yet. You can add them either in\n'
           '<code>File > Project Settings... > Add Image Files</code> or\n'
           '<code>File > Add New Image</code>.', UserWarning)
      return
    self._projImgMgr.show()
    self._projImgMgr.raise_()

  def addImages_gui(self):
    wiz = QtWidgets.QWizard()
    page = NewProjectWizard.createFilePage('Images', wiz)
    wiz.addPage(page)
    if wiz.exec_():
      for file in page.fileList.files:
        self.projData.addImageByPath(file)

  def addAnnotations_gui(self):
    wiz = QtWidgets.QWizard()
    page = NewProjectWizard.createFilePage('Annotations', wiz)
    wiz.addPage(page)
    if wiz.exec_():
      for file in page.fileList.files:
        self.projData.addAnnotationByPath(file)

  def create_gui(self):
    wiz = NewProjectWizard(self)
    ok = wiz.exec_()
    if not ok:
      return
    parsedLists = {}
    for k, val in wiz.fileLists.items():
      parsedLists[k] = val.files
    # Since insertion order is preserved the extraction can be done without keys
    images, annotations = parsedLists.values()
    settings = wiz.projSettings
    projName = settings['Name']
    prevTemplate = settings['Template Project']
    if prevTemplate is not None and len(prevTemplate) > 0:
      baseCfg = fns.attemptFileLoad(prevTemplate)
      if not settings['Keep Existing Images'] or 'images' not in baseCfg:
        baseCfg['images'] = []
      if not settings['Keep Existing Annotations'] or 'annotations' not in baseCfg:
        baseCfg['annotations'] = []
    else:
      baseCfg = {'images': [], 'annotations': []}
    baseCfg['images'].extend(images)
    baseCfg['annotations'].extend(annotations)
    projPath = Path(wiz.projSettings['Location'])/projName
    outPrj = self.projData.create(name=projPath, cfg=baseCfg)
    self.open(outPrj.cfgFname)


class NewProjectWizard(QtWidgets.QWizard):

  def __init__(self, project: FilePlugin, parent=None) -> None:
    super().__init__(parent)
    self.project = project
    self.fileLists : Dict[str, DropList] = {}

    # -----
    # PROJECT SETTINGS
    # -----
    page = QtWidgets.QWizardPage(self)
    page.setTitle('Project Settings')
    settings = [
      dict(name='Name', type='str', value='new-project'),
      dict(name='Location', type='filepicker', value='.', asFolder=True),
      dict(name='Template Project', type='filepicker', value=None,
           tip="Path to existing project config file. This will serve as a template for"
               " the newly created project, except for overridden settings"),
      dict(name='Keep Existing Images', type='bool', value=True,
           tip="Whether to keep images specified in the existing config"),
      dict(name='Keep Existing Annotations', type='bool', value=True,
           tip="Whether to keep annotations specified in the existing config"),
      # dict(name='Annotation Storage Format', value='pkl',
      #      tip="Project's internal representation of annotation files. Pickle (pkl) is the"
      #          " fastest for interaction, but is not as human readable. Alternatives (e.g. csv)"
      #          " are more human readable, but much slower when switching from image to image")
    ]
    # Use ParamEditor for speedy tree building
    editor = ParamEditor(paramList=settings, saveDir=None)
    tree = editor.tree
    self.projSettings = editor.params
    self.nameToPageMapping: Dict[str, QtWidgets.QWizardPage] = {}
    layout = QtWidgets.QVBoxLayout()
    page.setLayout(layout)
    layout.addWidget(tree)
    self.addPage(page)

    # -----
    # SELECTING PROJECT FILES
    # -----
    for fType in ['Images', 'Annotations']:
      page = self.createFilePage(fType, self)
      self.fileLists[fType] = page.fileList
      self.addPage(page)
      self.nameToPageMapping[fType] = page

  @staticmethod
  def createFilePage(name: str, wizard=None):
    page = QtWidgets.QWizardPage(wizard)
    page.setTitle(name)
    curLayout = QtWidgets.QVBoxLayout()
    page.setLayout(curLayout)
    curLayout.addWidget(QtWidgets.QLabel(f'New project {name.lower()} are shown below. Use the buttons'
                                         ' or drag and drop to add files.'))
    flist = DropList(wizard)
    page.fileList = flist
    fileBtnLayout = QtWidgets.QHBoxLayout()
    curLayout.addWidget(flist)
    for title in f'Add Files', f'Add Folder':
      selectFolder = 'Folder' in title
      btn = QtWidgets.QPushButton(title, wizard)
      btn.clicked.connect(partial(getFileList, wizard, flist, title, selectFolder))
      fileBtnLayout.addWidget(btn)
    curLayout.addLayout(fileBtnLayout)
    return page

def getFileList(wizard, _flist: DropList, _title: str, _selectFolder=False):
  dlg = QtWidgets.QFileDialog()
  dlg.setModal(True)
  getFn = lambda *args, **kwargs: dlg.getOpenFileNames(*args, **kwargs, options=dlg.DontUseNativeDialog)[0]
  if _selectFolder:
    getFn = lambda *args, **kwargs: [dlg.getExistingDirectory(*args, **kwargs, options=dlg.DontUseNativeDialog)]
  files = getFn(wizard, _title)
  _flist.addItems(files)

class ProjectImageManager(QtWidgets.QDockWidget):
  def __init__(self, parent=None):
    super().__init__(parent)
    self.setWindowTitle('Project Images')
    self.setObjectName('Project Images')
    self.setFeatures(self.DockWidgetMovable|self.DockWidgetFloatable|self.DockWidgetClosable)
    wid = QtWidgets.QWidget()
    self.setWidget(wid)

    layout = QtWidgets.QVBoxLayout()
    wid.setLayout(layout)

    self.thumbnails = ThumbnailViewer()
    self.completer = QtWidgets.QLineEdit()
    self.completer.setPlaceholderText('Type to filter')
    self.completer.textChanged.connect(self._filterThumbnails)

    self.sigImageSelected = self.thumbnails.sigImageSelected
    self.sigDeleteRequested = self.thumbnails.sigDeleteRequested

    layout.addWidget(self.completer)
    layout.addWidget(self.thumbnails)

  def _filterThumbnails(self, text):
    for ii in range(self.thumbnails.model().rowCount()):
      item = self.thumbnails.item(ii)
      if text in item.text():
        item.setHidden(False)
      else:
        item.setHidden(True)

class ProjectData(QtCore.QObject):
  sigCfgLoaded = QtCore.Signal()

  sigImagesAdded = QtCore.Signal(object)
  """List[Path] of added images"""
  sigImagesRemoved = QtCore.Signal(object)
  """List[Path] of removed images"""
  sigImagesMoved = QtCore.Signal(object)
  """
  List[(oldPath, NewPath)]
  Used mainly when images from outside the project are annotated. In that case, images are copied to inside the project,
  and this signal will be emitted.  
  """

  sigAnnotationsAdded = QtCore.Signal(object)
  """List[Path]"""
  sigAnnotationsRemoved = QtCore.Signal(object)
  """List[Path]"""


  def __init__(self, cfgFname: FilePath=None, cfgDict: dict=None):
    super().__init__()
    self.tableData = PRJ_SINGLETON.tableData
    self.templateName = PROJ_BASE_TEMPLATE
    self.cfg = fns.attemptFileLoad(self.templateName)
    self.cfgFname: Optional[Path] = None
    self.images: List[Path] = []
    self.baseImgDirs: Set[Path] = set()
    self.imgToAnnMapping: Dict[Path, Path] = {}
    """Records annotations belonging to each image"""

    self._suppressSignals = False
    """If this is *True*, no signals will be emitted """

    if cfgFname is not None or cfgDict is not None:
      self.loadCfg(cfgFname, cfgDict)

    self.compIo = ComponentIO()

  @property
  def location(self):
      return self.cfgFname.parent
  @property
  def imagesDir(self):
      return self.location/'images'
  @property
  def annotationsDir(self):
      return self.location/'annotations'
  @property
  def startup(self):
    return self.cfg['startup']
  @property
  def pluginCfg(self) -> Dict[str, str]:
      return self.cfg['plugin-cfg']

  def clearImgs_anns(self):
    oldImgs = self.images.copy()
    for lst in self.images, self.imgToAnnMapping, self.baseImgDirs:
      lst.clear()
    self._maybeEmit(self.sigImagesRemoved, oldImgs)

  def loadCfg(self, cfgFname: FilePath, cfgDict: dict = None, force=False):
    """
    Loads the specified project configuration by name (if a file) or dict (if programmatically created)
    :param cfgFname: Name of file to open. If `cfgDict` is provided instead, it will be saved here.
    :param cfgDict: If provided, this config is used and saved to `cfgFname` instead of using the file.
    :param force: If *True*, the new config will be loaded even if it is the same name as the
      current config
    """
    _, baseCfgDict = fns.resolveYamlDict(self.templateName)
    cfgFname, cfgDict = fns.resolveYamlDict(cfgFname, cfgDict)
    cfgFname = cfgFname.resolve()
    if not force and self.cfgFname == cfgFname:
      return None

    hierarchicalUpdate(baseCfgDict, cfgDict)

    loadPrjPlugins = baseCfgDict.get('plugin-cfg', {})
    newPlugins = {k: v for (k, v) in loadPrjPlugins.items() if k not in self.pluginCfg}
    if self.pluginCfg and newPlugins:
      raise ValueError('The previous project loaded custom plugins, which cannot easily'
                       ' be removed. To load a new project, close and re-open S3A with'
                       ' the new project instance instead.')

    self.cfgFname = cfgFname
    cfg = self.cfg = baseCfgDict
    self.annotationsDir.mkdir(exist_ok=True)
    self.imagesDir.mkdir(exist_ok=True)
    self.clearImgs_anns()

    warnPlgs = []
    for plgName, plgPath in newPlugins.items():
      pluginCls = pydoc.locate(plgPath)
      if pluginCls:
        PRJ_SINGLETON.addPlugin(pluginCls)
      else:
        warnPlgs.append(plgPath)
    if warnPlgs:
      warnLater(f'Some project plugins were specified, but could not be found:\n'
           f'{warnPlgs}', UserWarning)

    tableInfo = cfg.get('table-cfg', None)
    if isinstance(tableInfo, str):
      tableDict = None
      tableName = tableInfo
    else:
      if tableInfo is None:
        tableInfo = {}
      tableDict = tableInfo
      tableName = cfgFname
    tableName = Path(tableName)
    if not tableName.is_absolute():
      tableName = self.location/tableName
    self.tableData.loadCfg(tableName, tableDict, force=True)

    allAddedImages = []
    with self.suppressSignals():
      allAddedImages.extend(self.addImageFolder(self.imagesDir))
      for image in self.cfg['images']:
        if isinstance(image, dict):
          image.setdefault('copyToProj', False)
          renamed = self.addImage(**image)
          if renamed is not None:
            allAddedImages.append(renamed)
        else:
          allAddedImages.extend(self.addImageByPath(image, False))
    self._maybeEmit(self.sigImagesAdded, allAddedImages)

    with self.suppressSignals():
      self.addAnnotationFolder(self.annotationsDir)
      for annotation in set(self.cfg['annotations']) - {self.annotationsDir}:
        if isinstance(annotation, dict):
          self.addAnnotation(**annotation)
        else:
          self.addAnnotationByPath(annotation)
    self._maybeEmit(self.sigAnnotationsAdded, list(self.imgToAnnMapping.values()))

    self.sigCfgLoaded.emit()
    return self.cfgFname

  @classmethod
  def create(cls, *, name: FilePath= f'./{PROJ_FILE_TYPE}', cfg: dict=None, parent: ProjectData=None):
    """
    Creates a new project with the specified settings in the specified directory.
    :param name:
      helpText: Project Name. The parent directory of this name indicates the directory in which to create the project
      pType: filepicker
    :param cfg: see `ProjectData.loadCfg` for information
    :param parent: Associated ProjectData instance for a non-classmethod version of this function
    """
    name = Path(name)
    name = name/f'{name.name}.{PROJ_FILE_TYPE}'
    location = name.parent
    location.mkdir(exist_ok=True, parents=True)
    if parent is None:
      parent = cls()

    if not name.exists() and cfg is None:
      cfg = {}
    parent.loadCfg(name, cfg)

    tdName = Path(parent.tableData.cfgFname)
    if tdName.resolve() != parent.cfgFname:
      tdName = tdName.name
      fns.saveToFile(parent.tableData.cfg, location / tdName, True)
      parent.tableData.cfgFname = tdName
      parent.cfg['table-cfg'] = tdName

    parent.saveCfg()
    return parent

  def saveCfg(self):
    location = self.location
    annDir = self.annotationsDir
    strImgNames = []
    for folder in self.baseImgDirs:
      if location in folder.parents:
        folder = folder.relative_to(location)
      strImgNames.append(str(folder))
    for img in self.images:
      if img.parent in self.baseImgDirs:
        # This image is already accounted for in the base directories
        continue
      strImgNames.append(str(img.resolve()))

    offendingAnns = []
    for img, ann in self.imgToAnnMapping.items():
      if ann.parent != annDir:
        offendingAnns.append(str(ann))
        self.addAnnotation(ann)
    if len(offendingAnns) > 0:
      warnLater('Encountered annotation(s) in project config, but not officially added. '
            'Adding them now.'
           '  Offending files:\n'
           + ',\n'.join(offendingAnns), UserWarning)
    self.cfg['images'] = strImgNames
    # 'Ann' folder is always added on startup so no need to record it here. However,
    # if it is shown explicitly the user is aware.
    self.cfg['annotations'] = [self.annotationsDir.name]
    tblName = Path(self.tableData.cfgFname).absolute()
    if tblName !=  self.cfgFname:
      if tblName.parent == self.location:
        tblName = tblName.name
      self.cfg['table-cfg'] = str(tblName)
    fns.saveToFile(self.cfg, self.cfgFname)

  def addImageByPath(self, name: FilePath, copyToProj=False):
    """
    Determines whether to add as a folder or file based on filepath type. Since adding a folder returns a list of
    images and adding a single image returns a name or None, this function unifies the return signature by always
    returning a list. If the path is a single image and not a folder, and the return value is *None*, this function
    will return an empty list instead.
    """
    image = Path(name)
    if not image.is_absolute():
      image = self.location/image
    if not image.exists():
      warnLater(f'Provided image path does not exist: {image}\nNo action performed.',
            UserWarning)
      return []
    if image.is_dir():
      ret = self.addImageFolder(image, copyToProj)
    else:
      ret = self.addImage(name, copyToProj=copyToProj)
      ret = [] if ret is None else [ret]
    return ret

  def addImage(self, name: FilePath, data: NChanImg=None, copyToProj=False, allowOverwrite=False) -> Optional[FilePath]:
    fullName = Path(name)
    if not fullName.is_absolute():
      fullName = self.imagesDir / fullName
    if copyToProj or data is not None:
      fullName = self._copyImgToProj(fullName, data, allowOverwrite)
    if fullName.name in [i.name for i in self.images]:
      # Indicate the image was already present to calling scope
      return None
    self.images.append(fullName)
    self._maybeEmit(self.sigImagesAdded, [fullName])
    return fullName
    # TODO: Create less hazardous undo operation
    # yield name
    # self.removeImage(name)

  def changeImgPath(self, oldName: Path, newName: Path=None):
    oldIdx = self.images.index(oldName)
    if newName is None or newName in self.images:
      del self.images[oldIdx]
    else:
      self.images[oldIdx] = newName
    self._maybeEmit(self.sigImagesMoved, [(oldName, newName)])

  def addImageFolder(self, folder: FilePath, copyToProj=False):
    folder = Path(folder).resolve()
    if folder in self.baseImgDirs:
      return []
    if copyToProj:
      newFolder = self.imagesDir/folder.name
      shutil.copytree(folder, newFolder)
      folder = newFolder
      copyToProj = False
    self.baseImgDirs.add(folder)
    # Need to keep track of actually added images instead of using all globbed images. If an added image already
    # existed in the project, it won't be added. Also, if the images are copied into the project, the paths
    # will change.
    addedImgs = []
    with self.suppressSignals():
      for img in folder.glob('*.*'):
        finalName = self.addImage(img, copyToProj=copyToProj)
        if finalName is not None:
          addedImgs.append(finalName)
    self._maybeEmit(self.sigImagesAdded, addedImgs)

    return addedImgs


  def addAnnotationByPath(self, name: FilePath):
    """Determines whether to add as a folder or file based on filepath type"""
    name = Path(name)
    if not name.is_absolute():
      name = self.location/name
    if not name.exists():
      warnLater(f'Provided annotation path does not exist: {name}\nNo action '
                f'performed.', UserWarning)
      return
    if name.is_dir():
      self.addAnnotationFolder(name)
    else:
      self.addAnnotation(name)

  def addAnnotationFolder(self, folder: FilePath):
    folder = Path(folder).resolve()
    for file in folder.glob('*.*'):
      self.addAnnotation(file)

  def addImage_gui(self, copyToProject=True):
    fileFilter = "Image Files (*.png *.tif *.jpg *.jpeg *.bmp *.jfif);;All files(*.*)"
    fname = fns.popupFilePicker(None, 'Add Image to Project', fileFilter)
    if fname is not None:
      self.addImage(fname, copyToProj=copyToProject)

  def removeImage(self, imgName: FilePath):
    imgName = Path(imgName).resolve()
    if imgName not in self.images:
      return
    self.images.remove(imgName)
    # Remove copied annotations for this image
    for ann in self.annotationsDir.glob(f'{imgName.stem}.*'):
      ann.unlink()
    self.imgToAnnMapping.pop(imgName, None)
    self._maybeEmit(self.sigImagesRemoved, [imgName])
    if imgName.parent == self.imagesDir:
      imgName.unlink()
    # yield
    # if imgName.parent == self.imagesDir:
    #   # TODO: Cache removed images in a temp dir, then move them to that temp dir instead of unlinking
    #   #  on delete. This will make 'remove' undoable
    #   raise IOError('Can only undo undo image removal when the image was outside the project.'
    #                    f' Image {imgName.name} was either annotated or directly placed in the project images'
    #                    f' directory, and was deleted during removal. To re-add, do so from the original image'
    #                    f' location outside the project directory.')
    # self.addImage(imgName)

  def removeAnnotation(self, annName: FilePath):
    annName = Path(annName).resolve()
    # Since no mapping exists of all annotations, loop the long way until the file is found
    for key, ann in self.imgToAnnMapping.items():
      if annName == ann:
        del self.imgToAnnMapping[key]
        ann.unlink()
        break

  def addAnnotation(self, name: FilePath=None, data: pd.DataFrame=None, image: FilePath=None,
                    overwriteOld=False):
    # Housekeeping for default arguments
    if name is None and data is None:
      raise IOError('`name` and `data` cannot both be `None`')
    elif name in self.imgToAnnMapping.values() and not overwriteOld:
      # Already present, shouldn't be added
      return
    if data is None:
      data = ComponentIO.buildByFileType(name)
    if image is None:
      # If no explicit matching to an image is provided, try to determine based on annotation name
      xpondingImgs = np.unique(data[REQD_TBL_FIELDS.SRC_IMG_FILENAME].to_numpy())
      # Break into annotaitons by iamge
      for img in xpondingImgs:
        self.addAnnotation(name, data[data[REQD_TBL_FIELDS.SRC_IMG_FILENAME] == img], img)
      return
    image = self._getFullImgName(Path(image))
    # Since only one annotation file can exist per image, concatenate this with any existing files for the same image
    # if needed
    if image.parent != self.imagesDir:
      image = self._copyImgToProj(image)
    annForImg = self.imgToAnnMapping.get(image, None)
    oldAnns = []
    if annForImg is not None and not overwriteOld:
      oldAnns.append(ComponentIO.buildByFileType(annForImg))
    combinedAnns = oldAnns + [data]
    outAnn = pd.concat(combinedAnns, ignore_index=True)
    outAnn[REQD_TBL_FIELDS.INST_ID] = outAnn.index
    outFmt = f".{self.cfg['annotation-format']}"
    outName = self.annotationsDir / f'{image.name}{outFmt}'
    ComponentIO.exportByFileType(outAnn, outName, verifyIntegrity=False, readOnly=False, imgDir=self.imagesDir)
    self.imgToAnnMapping[image] = outName

  def _copyImgToProj(self, name: Path, data: NChanImg=None, overwrite=False):
    newName = self.imagesDir/name.name
    if newName.exists() and (not overwrite or newName == name):
      # Already in the project, no need to copy
      return newName
    if name.exists() and data is None:
      shutil.copy(name, newName)
    elif data is not None:
      # Programmatically created or not from a local file
      # noinspection PyTypeChecker
      cvImsave_rgb(newName, data)
    else:
      raise IOError(f'No image data associated with {name.name}. Either the file does not exist or no'
                       f' image information was provided.')
    if name in self.images:
      self.changeImgPath(name, newName)
      self._maybeEmit(self.sigImagesMoved, [(name, newName)])
    return newName

  def _getFullImgName(self, name: Path, thorough=True):
    """
    From an absolute or relative image name, attempts to find the absolute path it corresponds
    to based on current project images. A match is located in the following order:
      - If the image path is already absolute, it is resolved, checked for existence, and returned
      - Solitary project images are searched to see if they end with the specified relative path
      - All base image directories are checked to see if they contain this subpath

    :param thorough: If `False`, as soon as a match is found the function returns. Otherwise,
      all solitary paths and images will be checked to ensure there is exactly one matching
      image for the name provided.
    """
    if name.is_absolute():
      return name.resolve()

    candidates = set()
    strName = str(name)
    for img in self.images:
      if str(img).endswith(strName):
        if not thorough:
          return img
        candidates.add(img)

    for parent in self.baseImgDirs:
      curName = (parent/name).resolve()
      if curName.exists():
        if not thorough:
          return curName
        candidates.add(curName)

    numCandidates = len(candidates)
    if numCandidates != 1:
      msg = f'Exactly one corresponding image file must exist for a given annotation. However,' \
            f' {numCandidates} candidate images were found'
      if numCandidates == 0:
        msg += '.'
      else:
        msg += f':\n{", ".join([c.name for c in candidates])}'
      raise IOError(msg)
    return candidates.pop()

  def exportProj(self, outputFolder: FilePath= 's3a-export'):
    """
    Exports the entire project, making a copy of it at the destination directory
    :param outputFolder:
      helpText: Where to place the exported project
      pType: filepicker
      asFolder: True
    """
    shutil.copytree(self.location, outputFolder)

  @fns.dynamicDocstring(fileTypes=list(ComponentIO.handledIoTypes))
  def exportAnnotations(self, outputFolder:FilePath= 's3a-export',
                        annotationFormat='csv',
                        combine=False,
                        rescale=False,
                        includeImages=True):
    """
    Exports project annotations, optionally including their source images
    :param outputFolder:
      helpText: Folder for exported annotations
      pType: filepicker
      asFolder: True
    :param annotationFormat:
      helpText: "Annotation file type. E.g. if 'csv', annotations will be saved as csv files. Available
      file types are:
      {fileTypes}"
      pType: list
      limits:
        {fileTypes}
    :param combine: If `True`, all annotation files will be combined into one exported file with name `annotations.<format>`
    :param rescale:
      helpText: "For image exports, this determines whether the ouptut should retain the same
        values as their integer label or whether they should be scaled to entire range
        of black -> white"
    :param includeImages: If `True`, the corresponding image for each annotation will also be exported into an `images`
      folder
    """
    self.saveCfg()
    outputFolder = Path(outputFolder)

    if outputFolder.resolve() == self.cfgFname and not combine:
      return

    outputFolder.mkdir(exist_ok=True)
    if includeImages:
      outImgDir = outputFolder / 'images'
      outImgDir.mkdir(exist_ok=True)
      for img in self.imagesDir.glob('*.*'):
        if self.imgToAnnMapping.get(img, None) is not None:
          shutil.copy(img, outImgDir)

    ioArgs = {'imgDir': self.imagesDir, 'rescaleOutput': rescale}
    existingAnnFiles = [f for f in self.imgToAnnMapping.values() if f is not None]
    if combine:
      outAnn = pd.concat(map(self.compIo.buildByFileType, existingAnnFiles))
      self.compIo.exportByFileType(outAnn, outputFolder / f'annotations.{annotationFormat}', **ioArgs)
    else:
      outAnnsDir = outputFolder / 'annotations'
      outAnnsDir.mkdir(exist_ok=True)
      if self.cfg['annotation-format'] == annotationFormat:
        shutil.copytree(self.annotationsDir, outAnnsDir)
      else:
        for annFile in existingAnnFiles:
          self.compIo.convert(annFile, outAnnsDir/f'{annFile.stem}.{annotationFormat}', importArgs=ioArgs, exportArgs=ioArgs)

  def _maybeEmit(self, signal: QtCore.Signal, emitList: Sequence[Union[Path, Tuple[Path, Path]]]):
    if not self._suppressSignals:
      signal.emit(emitList)

  @contextmanager
  def suppressSignals(self):
    oldSuppress = self._suppressSignals
    self._suppressSignals = True
    yield
    self._suppressSignals = oldSuppress


# noinspection PyUnresolvedReferences
from utilitys.typeoverloads import FilePath

from numpy import ndarray
"""
Functions that just return a `np.ndarray` are often hard to interpret. Is the output a simple array, image, 
etc.? If it _is_ an image, what is the output shape? Developers and users will have to comb through the function body
or rely on detailed documentation for an answer. This is a simple answer to the problem. By redefining np.ndarray in 
several ways, users and devs can more clearly interpret the intenionality behind various np.ndarray (and other) types.
"""

BlackWhiteImg = ndarray
GrayImg = ndarray
RgbImg = ndarray
RgbaImg = ndarray
NChanImg = ndarray
OneDArr = ndarray
TwoDArr = ndarray
ThreeDArr = ndarray


import typing as t
from pathlib import Path
from typing import List, Dict, Union, Callable, Any

import numpy as np
import pandas as pd
from pyqtgraph.parametertree import Parameter
from utilitys import ParamEditor, fns

from s3a import PRJ_SINGLETON
from s3a.constants import APP_STATE_DIR
from s3a.generalutils import safeCallFuncList, hierarchicalUpdate, safeCallFunc
from s3a.structures import FilePath
from utilitys.fns import serAsFrame, attemptFileLoad, raiseErrorLater, warnLater


class AppStateEditor(ParamEditor):

  def __init__(self, parent=None, paramList: List[Dict] = None,
               saveDir: FilePath = APP_STATE_DIR, fileType='param', name=None,
               topTreeChild: Parameter = None):
    # TODO: Add params to choose which features are saved, etc.
    super().__init__(parent, paramList, saveDir, fileType, name, topTreeChild)
    self._stateFuncsDf = pd.DataFrame(columns=['importFuncs', 'exportFuncs', 'required'])
    self.loading = False

    self.startupSettings = {}

  def saveParamValues(self, saveName: str=None, paramState: dict=None, **kwargs):
    if saveName is None:
      saveName = self.RECENT_STATE_FNAME
    if paramState is None:
      # TODO: May be good in the future to be able to choose which should be saved
      legitKeys = self._stateFuncsDf.index
      exportFuncs = self._stateFuncsDf.exportFuncs
      saveOnExitDir = self.saveDir/'saved_on_exit'
      saveOnExitDir.mkdir(exist_ok=True)
      rets, errs = safeCallFuncList(legitKeys, exportFuncs, [[saveOnExitDir]] * len(legitKeys))
      updateDict = {k: ret for k, ret in zip(legitKeys, rets) if ret is not None}
      paramState = dict(Parameters=paramState, **updateDict)
      for editor in PRJ_SINGLETON.quickLoader.listModel.uniqueEditors:
        editor.applyChanges()
        if editor.lastAppliedName == 'Default':
          continue
        curSaveName = str(saveOnExitDir/editor.name)
        formattedName = editor.name.replace(' ', '').lower()
        editor.saveParamValues(curSaveName, blockWrite=False)
        paramState.update({formattedName: curSaveName})
    else:
      errs = []

    ret = super().saveParamValues(saveName, paramState, **kwargs)
    self.raiseErrMsgIfNeeded(errs)
    return ret

  def loadParamValues(self, stateName: Union[str, Path]=None, stateDict: dict = None, **kwargs):
    self.loading = True
    # Copy old settings to put them back after loading
    oldStartup = self.startupSettings.copy()
    try: # try block to ensure loading is false after
      if stateName is None:
        stateName = self.RECENT_STATE_FNAME
      if not stateName.exists() and stateDict is None:
        stateDict = {}
      if isinstance(stateDict, str):
        stateDict = {'quickloader': stateDict}
      stateDict = self._parseStateDict_includeRequired(stateName, stateDict)
      paramDict = stateDict.pop('Parameters', {}) or {}

      # It's possible for some functions (e.g. project load) to add or remove startup args,
      # so chack for this
      hierarchicalUpdate(self.startupSettings, kwargs)
      def nextKey():
        hierarchicalUpdate(stateDict, self.startupSettings)
        self.startupSettings.clear()
        legitKeys = self._stateFuncsDf.index.intersection(stateDict)
        if legitKeys.size > 0:
          return legitKeys[0]

      key = nextKey()
      rets, errs = [], {}
      while key:
        importFunc = self._stateFuncsDf.loc[key, 'importFuncs']
        arg = stateDict.pop(key, None)
        curRet, curErr = safeCallFunc(key, importFunc, arg)
        rets.append(curRet)
        if curErr: errs[key] = curErr
        key = nextKey()
      if errs:
        errPrint = [f'{k}: {v}' for (k, v) in errs.items()]
        warnLater('The following settings could not be loaded (shown as [setting]: [exception])\n'
             + "\n\n".join(errPrint), UserWarning)
      if stateDict:
        PRJ_SINGLETON.quickLoader.buildFromStartupParams(stateDict)
      ret = super().loadParamValues(stateName, paramDict)
    finally:
      self.loading = False
      hierarchicalUpdate(self.startupSettings, oldStartup)
    return ret

  def _parseStateDict_includeRequired(self, stateName: t.Union[str, Path], stateDict: dict = None):
    if self.RECENT_STATE_FNAME.exists():
      defaults = attemptFileLoad(self.RECENT_STATE_FNAME)
    else:
      defaults = {}
    try:
      out = self._parseStateDict(stateName, stateDict)
    except FileNotFoundError:
      out = {}
    for k in self._stateFuncsDf.index[self._stateFuncsDf['required']]:
      out.setdefault(k, defaults.get(k))
    return out

  @staticmethod
  def raiseErrMsgIfNeeded(errMsgs: List[str]):
    if len(errMsgs) > 0:
      err = IOError('Errors were encountered for the following parameters'
                         ' (shown as [parameter]: [exception])\n'
                       + "\n\n".join(errMsgs))
      fns.raiseErrorLater(err)


  def addImportExportOpts(self, optName: str, importFunc: Callable[[str], Any],
                          exportFunc: Callable[[Path], str], index:int=None,
                          required=False):
    """
    Main interface to the app state editor. By providing import and export functions,
    various aspects of the program state can be loaded and saved on demand.

    :param optName: What should this save option be called? E.g. when providing a
      load and save for annotation data, this is 'annotations'.
    :param importFunc: Function called when importing saved data. Takes in a
      full file path
    :param exportFunc: Function to save the app data. Input is a full folder path. Expects
      the output to be a full file path of the saved file. This file is then passed to
      'importFunc' on loading a param state
    :param index: Where to place this function. In most cases, this won't matter. However, some imports must be
      performed first / last otherwise app behavior may be undefined. In these cases, passing a value for index ensures
      correct placement of the import/export pair. By default, the function is added to the end of the import/export list.
    :param required: If *True*, this parameter is required every time param values are loaded.
      In the case it is missing from a load, the param editor first attempts to fetch this option
      from the most recent saved state.
    """
    newRow = pd.Series([importFunc, exportFunc, required], name=optName,
                       index=self._stateFuncsDf.columns)
    if index is not None:
      # First, shift old entries
      df = self._stateFuncsDf
      self._stateFuncsDf = pd.concat([df.iloc[:index], serAsFrame(newRow), df.iloc[index:]])
    else:
      self._stateFuncsDf: pd.DataFrame = self._stateFuncsDf.append(newRow)

  @property
  def RECENT_STATE_FNAME(self):
      return self.saveDir/f'recent.{self.fileType}'



from typing import Union, Any
from warnings import warn

import numpy as np
import pandas as pd
import cv2 as cv
from pandas import DataFrame as df
from pyqtgraph.Qt import QtCore

from s3a import PRJ_SINGLETON
from s3a.generalutils import coerceDfTypes
from s3a.constants import REQD_TBL_FIELDS as RTF, PRJ_ENUMS
from s3a.structures import ComplexXYVertices, OneDArr

__all__ = ['ComponentMgr', 'CompTableModel']

from utilitys.fns import warnLater

Signal = QtCore.Signal

TBL_FIELDS = PRJ_SINGLETON.tableData.allFields

class CompTableModel(QtCore.QAbstractTableModel):
  # Emits 3-element dict: Deleted comp ids, changed comp ids, added comp ids
  defaultEmitDict = {'deleted': np.array([]), 'changed': np.array([]), 'added': np.array([])}
  sigCompsChanged = Signal(dict)
  sigFieldsChanged = Signal()

  # Used for efficient deletion, where deleting non-contiguous rows takes 1 operation
  # Instead of N operations

  def __init__(self):
    super().__init__()
    # Create component dataframe and remove created row. This is to
    # ensure datatypes are correct
    self.resetFields()

  # ------
  # Functions required to implement table model
  # ------
  def columnCount(self, *args, **kwargs):
    return len(self.colTitles)

  def rowCount(self, *args, **kwargs):
    return len(self.compDf)

  def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
    if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
      return self.colTitles[section]

  # noinspection PyMethodOverriding
  def data(self, index: QtCore.QModelIndex, role: int) -> Any:
    outData = self.compDf.iloc[index.row(), index.column()]
    if role == QtCore.Qt.DisplayRole:
      return str(outData)
    elif role == QtCore.Qt.EditRole:
      return outData
    else:
      return None

  @PRJ_SINGLETON.actionStack.undoable('Alter Component Data')
  def setData(self, index, value, role=QtCore.Qt.EditRole) -> bool:
    row = index.row()
    col = index.column()
    oldVal = self.compDf.iat[row, col]
    # Try-catch for case of numpy arrays
    noChange = oldVal == value
    try:
      if noChange:
        return True
    except ValueError:
      # Happens with array comparison
      pass
    self.compDf.iat[row, col] = value
    # !!! Serious issue! Using iat sometimes doesn't work and I have no idea why since it is
    # not easy to replicate. See https://github.com/pandas-dev/pandas/issues/22740
    # Also, pandas iloc unnecessarily coerces to 2D ndarray when setting, so iloc will fail
    # when assigning an array to a single location. Not sure how to prevent this...
    # For now, checking this on export
    cmp = self.compDf.iloc[row, [col, col-1]].values[0] != self.compDf.iat[row, col]
    try:
      cmp = bool(cmp)
    except ValueError:
      # Numpy array-like
      cmp = np.any(cmp)
    if cmp:
      warnLater('Warning! An error occurred setting this value. Please try again using a'
           ' <em>multi-cell</em> edit. E.g. do not just set this value, set it along with'
           ' at least one other selected cell.', UserWarning)
    toEmit = self.defaultEmitDict.copy()
    toEmit['changed'] = np.array([self.compDf.index[index.row()]])
    self.sigCompsChanged.emit(toEmit)
    yield True
    self.setData(index, oldVal, role)
    return True

  def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlags:
    if index.column() not in self.noEditColIdxs:
      return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable
    else:
      return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

  # noinspection PyAttributeOutsideInit
  def resetFields(self):
    self.colTitles = [f.name for f in TBL_FIELDS]

    self.compDf = PRJ_SINGLETON.tableData.makeCompDf(0)

    noEditParams = [f for f in PRJ_SINGLETON.tableData.allFields if f.opts.get('readonly', False)]

    self.noEditColIdxs = [self.colTitles.index(col.name) for col in noEditParams]
    self.editColIdxs = np.setdiff1d(np.arange(len(self.colTitles)), self.noEditColIdxs)
    self.sigFieldsChanged.emit()

class ComponentMgr(CompTableModel):
  _nextCompId = 0
  compDf: pd.DataFrame

  def resetFields(self):
    super().resetFields()
    self._nextCompId = 0

  @PRJ_SINGLETON.actionStack.undoable('Add Components')
  def addComps(self, newCompsDf: df, addtype: PRJ_ENUMS = PRJ_ENUMS.COMP_ADD_AS_NEW, emitChange=True):
    toEmit = self.defaultEmitDict.copy()
    existingIds = self.compDf.index

    if len(newCompsDf) == 0:
      # Nothing to undo
      return toEmit

    # Delete entries with no vertices, since they make work within the app difficult.
    # TODO: Is this the appropriate response?
    verts = newCompsDf[RTF.VERTICES]
    dropIds = newCompsDf.index[verts.map(ComplexXYVertices.isEmpty)]
    newCompsDf.drop(index=dropIds, inplace=True)

    if addtype == PRJ_ENUMS.COMP_ADD_AS_NEW:
      # Treat all comps as new -> set their IDs to guaranteed new values
      newIds = np.arange(self._nextCompId, self._nextCompId + len(newCompsDf), dtype=int)
      newCompsDf[RTF.INST_ID] = newIds
      newCompsDf.set_index(newIds, inplace=True)
      dropIds = np.array([], dtype=int)
    else:
      # Merge may have been performed with new comps (id -1) mixed in
      needsUpdatedId = newCompsDf.index == RTF.INST_ID.value
      newIds = np.arange(self._nextCompId, self._nextCompId + np.sum(needsUpdatedId), dtype=int)
      newCompsDf.loc[needsUpdatedId, RTF.INST_ID] = newIds

    # Track dropped data for undo
    alteredIdxs = np.concatenate([newCompsDf.index.values, dropIds])
    alteredDataDf = self.compDf.loc[self.compDf.index.intersection(alteredIdxs)]

    # Delete entries that were updated to have no vertices
    toEmit.update(self.rmComps(dropIds, emitChange=False))
    # Now, merge existing IDs and add new ones
    newIds = newCompsDf.index
    newChangedIdxs = np.isin(newIds, existingIds, assume_unique=True)
    changedIds = newIds[newChangedIdxs]

    # Signal to table that rows should change
    self.layoutAboutToBeChanged.emit()
    # Ensure indices overlap with the components these are replacing
    self.compDf.update(newCompsDf)
    toEmit['changed'] = changedIds

    # Finally, add new comps
    compsToAdd = newCompsDf.iloc[~newChangedIdxs, :]
    # Make sure all required data is present for new rows
    missingCols = np.setdiff1d(self.compDf.columns, compsToAdd.columns)
    if missingCols.size > 0 and len(compsToAdd) > 0:
      embedInfo = PRJ_SINGLETON.tableData.makeCompDf(len(newCompsDf)).set_index(compsToAdd.index)
      compsToAdd[missingCols] = embedInfo[missingCols]
    self.compDf = pd.concat((self.compDf, compsToAdd), sort=False)
    # Retain type information
    coerceDfTypes(self.compDf)

    toEmit['added'] = newIds[~newChangedIdxs]
    self.layoutChanged.emit()


    self._nextCompId = np.max(self.compDf.index.to_numpy(), initial=-1) + 1

    if emitChange:
      self.sigCompsChanged.emit(toEmit)

    yield toEmit

    # Undo add by deleting new components and un-updating existing ones
    self.addComps(alteredDataDf, PRJ_ENUMS.COMP_ADD_AS_MERGE)
    addedCompIdxs = toEmit['added']
    if len(addedCompIdxs) > 0:
      self.rmComps(toEmit['added'])

  @PRJ_SINGLETON.actionStack.undoable('Remove Components')
  def rmComps(self, idsToRemove: Union[np.ndarray, type(PRJ_ENUMS)] = PRJ_ENUMS.COMP_RM_ALL,
              emitChange=True) -> dict:
    toEmit = self.defaultEmitDict.copy()
    # Generate ID list
    existingCompIds = self.compDf.index
    if idsToRemove is PRJ_ENUMS.COMP_RM_ALL:
      idsToRemove = existingCompIds
    elif not hasattr(idsToRemove, '__iter__'):
      # single number passed in
      idsToRemove = [idsToRemove]
      pass
    idsToRemove = np.array(idsToRemove)

    # Do nothing for IDs not actually in the existing list
    idsActuallyRemoved = np.isin(idsToRemove, existingCompIds, assume_unique=True)
    if len(idsActuallyRemoved) == 0:
      return toEmit
    idsToRemove = idsToRemove[idsActuallyRemoved]

    # Track for undo purposes
    removedData = self.compDf.loc[idsToRemove]

    tfKeepIdx = np.isin(existingCompIds, idsToRemove, assume_unique=True, invert=True)

    # Reset manager's component list
    self.layoutAboutToBeChanged.emit()
    self.compDf: df = self.compDf.iloc[tfKeepIdx,:]
    self.layoutChanged.emit()

    # Preserve type information after change
    coerceDfTypes(self.compDf)

    # Determine next ID for new components
    self._nextCompId = 0
    if np.any(tfKeepIdx):
      self._nextCompId = np.max(existingCompIds[tfKeepIdx].to_numpy()) + 1

    # Reflect these changes to the component list
    toEmit['deleted'] = idsToRemove
    if emitChange:
      self.sigCompsChanged.emit(toEmit)
    if len(idsToRemove) > 0:
      yield toEmit
    else:
      # Nothing to undo
      return toEmit

    # Undo code
    self.addComps(removedData, PRJ_ENUMS.COMP_ADD_AS_MERGE)

  @PRJ_SINGLETON.actionStack.undoable('Merge Components')
  def mergeCompVertsById(self, mergeIds: OneDArr=None, keepId: int=None):
    """
    Merges the selected components

    :param mergeIds: Ids of components to merge. If *None*, defaults to current user
      selection.
    :param keepId: If provided, the selected component with this ID is used as
      the merged component columns (except for the vertices, of course). Else,
      this will default to the first component in the selection.
    """
    if mergeIds is None or len(mergeIds) < 2:
      warn(f'Less than two components are selected, so "merge" is a no-op.', UserWarning)
      return
    mergeComps: df = self.compDf.loc[mergeIds].copy()
    if keepId is None:
      keepId = mergeIds[0]

    keepInfo = mergeComps.loc[keepId].copy()
    allVerts = [v.stack() for v in mergeComps[RTF.VERTICES]]
    maskShape = np.max(np.vstack(allVerts), 0)[::-1]
    mask = np.zeros(maskShape, bool)
    for verts in mergeComps[RTF.VERTICES]: # type: ComplexXYVertices
      mask |= verts.toMask(tuple(maskShape))
    newVerts = ComplexXYVertices.fromBwMask(mask)
    keepInfo[RTF.VERTICES] = newVerts

    self.rmComps(mergeComps.index)
    self.addComps(keepInfo.to_frame().T, PRJ_ENUMS.COMP_ADD_AS_MERGE)
    yield
    self.addComps(mergeComps, PRJ_ENUMS.COMP_ADD_AS_MERGE)

  @PRJ_SINGLETON.actionStack.undoable('Split Components')
  def splitCompVertsById(self, splitIds: OneDArr):
    """
    Makes a separate component for each distinct boundary in all selected components.
    For instance, if two components are selected, and each has two separate circles as
    vertices, then 4 total components will exist after this operation.

    Each new component will have the table fields of its parent

    :param splitIds: Ids of components to split up
    """
    splitComps = self.compDf.loc[splitIds].copy()
    newComps_lst = []
    for _, comp in splitComps.iterrows():
      verts: ComplexXYVertices = comp[RTF.VERTICES]
      tmpMask = verts.toMask(asBool=False).astype('uint8')
      nComps, ccompImg = cv.connectedComponents(tmpMask)
      newVerts = []
      for ii in range(1, nComps):
        newVerts.append(ComplexXYVertices.fromBwMask(ccompImg == ii))
      childComps = pd.concat([comp.to_frame().T]*(nComps-1))
      childComps[RTF.VERTICES] = newVerts
      newComps_lst.append(childComps)
    newComps = pd.concat(newComps_lst)
    # Keep track of which comps were removed and added by this op
    outDict = self.rmComps(splitComps.index)
    outDict.update(self.addComps(newComps))
    yield outDict
    undoDict = self.rmComps(newComps.index)
    undoDict.update(self.addComps(splitComps, PRJ_ENUMS.COMP_ADD_AS_MERGE))
    return undoDict

  def removeOverlapById(self, overlapIds: OneDArr):
    """
    Makes sure all specified components have no overlap. Preference is given
    in order of the given IDs, i.e. the last ID in the list is guaranteed to
    keep its full shape. If an area selection is made, priority is given to larger
    IDs, i.e. the largest ID is guaranteed to keep its full original shape.
    """
    overlapComps = self.compDf.loc[overlapIds].copy()
    allVerts = np.vstack([v.stack() for v in overlapComps[RTF.VERTICES]])
    wholeMask = np.zeros(allVerts.max(0)[::-1], dtype='uint16')
    for ii, (_, comp) in enumerate(overlapComps.iterrows(), 1):
      comp[RTF.VERTICES].toMask(wholeMask, ii, asBool=False)
    for ii, compId in enumerate(overlapIds, 1):
      verts = ComplexXYVertices.fromBwMask(wholeMask == ii)
      overlapComps.at[compId, RTF.VERTICES] = verts
    self.addComps(overlapComps, PRJ_ENUMS.COMP_ADD_AS_MERGE)


from __future__ import annotations

from typing import Callable, Sequence

import cv2 as cv
import numpy as np
import pandas as pd
from pyqtgraph import console as pg_console
from pyqtgraph.Qt import QtCore, QtWidgets, QtGui
from utilitys import ParamEditorPlugin, ProcessIO, widgets as uw

from s3a import models, XYVertices, ComplexXYVertices
from s3a.constants import PRJ_CONSTS as CNST, REQD_TBL_FIELDS as RTF, PRJ_ENUMS
from s3a.models import s3abase
from s3a.parameditors import PRJ_SINGLETON
from s3a.plugins.base import ProcessorPlugin


class MainImagePlugin(ParamEditorPlugin):
  name = __groupingName__ = 'Main Image'
  _makeMenuShortcuts = False

  def __init__(self):
    super().__init__()

  def attachWinRef(self, win: s3abase.S3ABase):
    mainImg = win.mainImg

    copier = mainImg.regionCopier
    def startCopy():
      """
      Copies the selected components. They can be pasted by <b>double-clicking</b>
      on the destination location. When done copying, Click the *Clear ROI* tool change
      the current draw action.
      """
      copier.inCopyMode = True
      copier.sigCopyStarted.emit()
    def startMove():
      """
      Moves the selected components. They can be pasted by <b>double-clicking</b>
      on the destination location.
      """
      copier.inCopyMode = False
      copier.sigCopyStarted.emit()

    self.registerFunc(startMove, btnOpts=CNST.TOOL_MOVE_REGIONS)
    self.registerFunc(startCopy, btnOpts=CNST.TOOL_COPY_REGIONS)
    copier.sigCopyStopped.connect(win.mainImg.updateFocusedComp)


    disp = win.compDisplay
    # Wrap in process to ignore the default param

    def actHandler(verts, param):
      # When editing, only want to select if nothing is already selected
      if (param not in [CNST.DRAW_ACT_REM, CNST.DRAW_ACT_ADD]
          or  len(disp.selectedIds) == 0
      ):
        # Special case: Selection with point shape should be a point
        if self.win.mainImg.shapeCollection.curShapeParam == CNST.DRAW_SHAPE_POINT:
          verts = verts.mean(0, keepdims=True)
        disp.reflectSelectionBoundsMade(verts)

    self.registerFunc(disp.mergeSelectedComps, btnOpts=CNST.TOOL_MERGE_COMPS, ignoreKeys=['keepId'])
    self.registerFunc(disp.splitSelectedComps, btnOpts=CNST.TOOL_SPLIT_COMPS)
    win.mainImg.registerDrawAction([CNST.DRAW_ACT_ADD, CNST.DRAW_ACT_REM, CNST.DRAW_ACT_SELECT, CNST.DRAW_ACT_PAN],
                                 actHandler)
    self.registerFunc(disp.removeSelectedCompOverlap, btnOpts=CNST.TOOL_REM_OVERLAP)

    win.mainImg.registerDrawAction(CNST.DRAW_ACT_CREATE, self.createComponent)
    win.mainImg.addTools(self.toolsEditor)

    # No need for a dropdown menu
    self.dock = None
    super().attachWinRef(win)

  @property
  def image(self):
    return self.win.mainImg.image

  def createComponent(self, roiVerts: XYVertices):
    verts = np.clip(roiVerts.astype(int), 0, self.image.shape[:2][::-1])

    if cv.contourArea(verts) < self.win.mainImg.minCompSize:
      # Use as selection instead of creation
      self.win.compDisplay.reflectSelectionBoundsMade(roiVerts[[0]])
      return

    # noinspection PyTypeChecker
    verts = ComplexXYVertices([verts])
    newComps = PRJ_SINGLETON.tableData.makeCompDf()
    newComps[RTF.VERTICES] = [verts]
    self.win.add_focusComps(newComps)

class CompTablePlugin(ParamEditorPlugin):
  name = 'Component Table'

  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()
    cls.dock.addEditors([PRJ_SINGLETON.filter])


  def attachWinRef(self, win: s3abase.S3ABase):

    tbl = win.compTbl
    for func, param in zip(
        [lambda: tbl.setSelectedCellsAs_gui(), tbl.removeSelectedRows_gui, tbl.setSelectedCellsAsFirst,
         lambda: win.compDisplay.scaleViewboxToSelectedIds()],
        [CNST.TOOL_TBL_SET_AS, CNST.TOOL_TBL_DEL_ROWS, CNST.TOOL_TBL_SET_SAME_AS_FIRST,
         CNST.TOOL_TBL_ZOOM_TO_COMPS]):
      param.opts['ownerObj'] = win.compTbl
      self.registerFunc(func, name=param.name, btnOpts=param)
    tbl.menu = self.toolsEditor.actionsMenuFromProcs(parent=tbl, nest=True)
    super().attachWinRef(win)

class EditPlugin(ParamEditorPlugin):

  name = '&Edit'

  def attachWinRef(self, win: s3abase.S3ABase):
    super().attachWinRef(win)
    stack = PRJ_SINGLETON.actionStack

    for param in CNST.TOOL_UNDO, CNST.TOOL_REDO: param.opts['ownerObj'] = win
    self.registerFunc(stack.undo, name='Undo', btnOpts=CNST.TOOL_UNDO)
    self.registerFunc(stack.redo, name='Redo', btnOpts=CNST.TOOL_REDO)

    def updateUndoRedoTxts():
      self.undoAct.setText(f'Undo: {stack.undoDescr}')
      self.redoAct.setText(f'Redo: {stack.redoDescr}')
    stack.stackChangedCallbacks.append(updateUndoRedoTxts)
    updateUndoRedoTxts()

  @property
  def undoAct(self):
      return [a for a in self.menu.actions() if a.text().startswith('Undo')][0]

  @property
  def redoAct(self):
    return [a for a in self.menu.actions() if a.text().startswith('Redo')][0]

class RandomToolsPlugin(ParamEditorPlugin):
  name = '&Tools'
  _showFuncDetails = True

  def attachWinRef(self, win: s3abase.S3ABase):
    super().attachWinRef(win)

    self.registerFunc(self.showDevConsole)
    self.registerFunc(win.clearBoundaries, btnOpts=CNST.TOOL_CLEAR_BOUNDARIES)
    self.registerFunc(win.compDisplay.exportCompOverlay, name='Export Component Overlay', toClipboard=True)
    self.registerFunc(lambda: win.setMainImg(None), name='Clear Current Image')

  def showDevConsole(self):
    """
    Opens a console that allows dynamic interaction with current variables. If IPython
    is on your system, a qt console will be loaded. Otherwise, a (less capable) standard
    pyqtgraph console will be used.
    """
    namespace = dict(app=self.win, rtf=RTF, singleton=PRJ_SINGLETON)
    # "dict" default is to use repr instead of string for internal elements, so expanding
    # into string here ensures repr is not used
    nsPrintout = [f"{k}: {v}" for k, v in namespace.items()]
    text = f'Starting console with variables:\n' \
           f'{nsPrintout}'
    # Broad exception is fine, fallback is good enough. Too many edge cases to properly diagnose when Pycharm's event
    # loop is sync-able with the Jupyter dev console
    # noinspection PyBroadException
    try:
      console = uw.ConsoleWidget(parent=self.win, namespace=namespace, text=text)
    except Exception:
      # Ipy kernel can have issues for many different reasons. Always be ready to fall back to traditional console
      console = pg_console.ConsoleWidget(parent=self.win, namespace=namespace, text=text)
    console.setWindowFlags(QtCore.Qt.Window)
    console.show()

class HelpPlugin(ParamEditorPlugin):
  name = '&Help'

  def attachWinRef(self, win: QtWidgets.QMainWindow):
    super().attachWinRef(win)
    self.registerFunc(lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl('https://gitlab.com/ficsresearch/s3a/-/wikis/home')),
                         name='Online User Guide')
    self.registerFunc(lambda: QtWidgets.QMessageBox.aboutQt(win, 'About Qt'), name='About Qt')
    self.registerFunc(self.iconAttributions)

  def iconAttributions(self):
    htmlStr = """
    <div>Icons made by <a href="https://www.freepik.com" title="Freepik">Freepik</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a></div>

    <div>Icons made by <a href="https://www.flaticon.com/authors/those-icons" title="Those Icons">Those Icons</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a></div>

    <div>Icons made by <a href="https://www.flaticon.com/authors/pixel-perfect" title="Pixel perfect">Pixel perfect</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a></div>
    """

    QtWidgets.QMessageBox.information(self.win, 'Icon Attributions', htmlStr)


def miscFuncsPluginFactory(name_: str=None, regFuncs: Sequence[Callable]=None, titles: Sequence[str]=None, showFuncDetails=False):
  class FuncContainerPlugin(ParamEditorPlugin):
    name = name_
    _showFuncDetails = showFuncDetails

    def attachWinRef(self, win: s3abase.S3ABase):
      super().attachWinRef(win)

      nonlocal regFuncs, titles
      if regFuncs is None:
        regFuncs = []
      if titles is None:
        titles = [None] * len(regFuncs)
      for func, title in zip(regFuncs, titles):
        self.registerFunc(func, title)

  return FuncContainerPlugin


class MultiPredictionsPlugin(ProcessorPlugin):
  name = 'Multi-Predictions'

  mgr: models.tablemodel.ComponentMgr

  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()
    cls.procCollection = PRJ_SINGLETON.multiPredClctn.createProcessorForClass(cls, cls.name + ' Processor')
    cls.dock.addEditors([cls.procCollection])

  def __init__(self):
    super().__init__()
    self.registerFunc(self.lastRunAnalytics)

  def attachWinRef(self, win):
    super().attachWinRef(win)
    self.mgr = win.compMgr
    self.mainImg = win.mainImg
    win.mainImg.toolsEditor.registerFunc(self.predictFromSelection, btnOpts=CNST.TOOL_PRED_SEL)

  def makePrediction(self, comps: pd.DataFrame):
    if self.win.mainImg.image is None:
      return
    vbRange = np.array(self.mainImg.getViewBox().viewRange()).T
    newComps = self.curProcessor.run(components=comps, image=self.win.mainImg.image,
                                     viewbox=vbRange)
    if not isinstance(newComps, ProcessIO):
      newComps = ProcessIO(components=newComps)
    compsToAdd = newComps['components']
    addType = newComps.get('addType', PRJ_ENUMS.COMP_ADD_AS_NEW)
    if newComps.get('deleteOrig', False):
      self.mgr.rmComps(comps.index)
    self.mgr.addComps(compsToAdd, addType)


  def predictFromSelection(self):
    self.makePrediction(self.mgr.compDf.loc[self.win.compDisplay.selectedIds])

  def lastRunAnalytics(self):
    raise NotImplementedError


from __future__ import annotations

from typing import Union

import pandas as pd
from utilitys import ParamEditorPlugin, NestedProcWrapper

from s3a import parameditors as pe
from s3a.constants import PRJ_CONSTS


class ProcessorPlugin(ParamEditorPlugin):
  procCollection: pe.algcollection.AlgParamEditor = None
  """
  Most table field plugins will use some sort of processor to infer field data.
  This property holds spawned collections. See :class:`XYVerticesPlugin` for
  an example.
  """

  @property
  def curProcessor(self):
    return self.procCollection.curProcessor

  @curProcessor.setter
  def curProcessor(self, newProcessor: Union[str, NestedProcWrapper]):
    self.procCollection.switchActiveProcessor(newProcessor)

class TableFieldPlugin(ProcessorPlugin):
  mainImg = None
  """
  Holds a reference to the focused image and set when the s3a reference is set. This
  is useful for most table field plugins, since mainImg will hold a reference to the
  component series that is modified by the plugins.
  """

  _active=False

  _makeMenuShortcuts = False

  @property
  def parentMenu(self):
    return self.win.tblFieldToolbar

  def attachWinRef(self, win):
    super().attachWinRef(win)
    self.mainImg = mainImg = win.mainImg
    win.sigRegionAccepted.connect(self.acceptChanges)
    mainImg.sigUpdatedFocusedComp.connect(self.updateFocusedComp)
    self.active = True
    self.registerFunc(self.processorAnalytics, btnOpts=PRJ_CONSTS.TOOL_PROC_ANALYTICS)

  def processorAnalytics(self):
    proc = self.curProcessor
    try:
      proc.processor.stageSummary_gui()
    except NotImplementedError:
      raise TypeError(f'Processor type {type(proc)} does not implement summary analytics.')


  def updateFocusedComp(self, newComp: pd.Series = None):
    """
    This function is called when a new component is created or the focused image is updated
    from the main view. See :meth:`MainImage.updateFocusedComp` for parameters.
    """
    pass

  def acceptChanges(self):
    """
    This must be overloaded by each plugin so the set component data is properly stored
    in the focused component. Essentially, any changes made by this plugin are saved
    after a call to this method.
    """
    raise NotImplementedError

  @property
  def active(self):
    """Whether this plugin is currently in use by the focused image."""
    return self._active

  @active.setter
  def active(self, newActive: bool):
    if newActive == self._active:
      return
    if newActive:
      self._onActivate()
    else:
      self._onDeactivate()
    self._active = newActive

  def _onActivate(self):
    """Overloaded by plugin classes to set up the plugin for use"""

  def _onDeactivate(self):
    """Overloaded by plugin classes to tear down when the plugin is no longer in use"""


import pickle
import sys
from ast import literal_eval
from collections import defaultdict
from pathlib import Path
from stat import S_IREAD, S_IRGRP, S_IROTH
from typing import Any, Optional, Union, Tuple, Callable

import numpy as np
import pandas as pd
from pandas import DataFrame as df
from skimage import io, measure
from skimage.exposure import rescale_intensity
from typing_extensions import Literal

from utilitys.fns import warnLater

from s3a.constants import REQD_TBL_FIELDS as RTF
from s3a.generalutils import augmentException, getCroppedImg, resize_pad, cvImsave_rgb
from s3a.parameditors.table import TableData
from s3a.structures import PrjParamGroup, FilePath, GrayImg, \
  ComplexXYVertices, PrjParam, XYVertices

FilePathOrDf = Union[FilePath, pd.DataFrame]

def _strSerToParamSer(strSeries: pd.Series, paramVal: Any) -> pd.Series:
  paramType = type(paramVal)
  # TODO: Move this to a more obvious place?
  funcMap = {
    # Format string to look like a list, use ast to convert that string INTO a list, make a numpy array from the list
    np.ndarray        : lambda strVal: np.array(literal_eval(strVal)),
    ComplexXYVertices : ComplexXYVertices.deserialize,
    XYVertices        : XYVertices.deserialize,
    bool              : lambda strVal: strVal.lower() == 'true',
    PrjParam          : lambda strVal: PrjParamGroup.fieldFromParam(paramVal.group, strVal)
  }
  defaultFunc = lambda strVal: paramType(strVal)
  funcToUse = funcMap.get(paramType, defaultFunc)
  return strSeries.apply(funcToUse).values


def _paramSerToStrSer(paramSer: pd.Series, paramVal: Any) -> pd.Series:
  # TODO: Move along with above function?
  paramType = type(paramVal)
  funcMap = {
    # Format string to look like a list, use ast to convert that string INTO a list, make a numpy array from the list
    np.ndarray: lambda param: str(param.tolist()),
    XYVertices: XYVertices.serialize,
    ComplexXYVertices: ComplexXYVertices.serialize,
  }
  defaultFunc = lambda param: str(param)

  funcToUse = funcMap.get(paramType, defaultFunc)
  return paramSer.apply(funcToUse)

class ComponentIO:
  """
  Exporter responsible for saving Component information to a file or object.
  Once created, users can extract different representations of components by
  calling exporter.exportCsv, exportPkl, etc. for those objects / files respectively.
  """
  handledIoTypes = {'csv': 'CSV Files', 'pkl': 'Pickle Files',
                    'id.png': 'ID Grayscale Image'}
  """Dict of <type, description> for the file types this I/O obejct can handle"""

  # Dictionary comprehension doesn't work in a class scope
  roundTripIoTypes = {}
  """
  Not all IO types can export->import and remain the exact same dataframe afterwards.
  For instance, exporting a labeled image will discard all additional fields.
  This property holds export types which can give back the original dataframe after
  a round trip export->import.
  """
  for k in ['csv', 'pkl']:
    roundTripIoTypes[k] = handledIoTypes[k]

  tableData = TableData()
  """Table to use for import/export cross checks. This is how class and table field information is derived."""

  @classmethod
  def handledIoTypes_fileFilter(cls, typeFilter='', **extraOpts):
    """
    Helper for creating a file filter out of the handled IO types. The returned list of
    strings is suitable for inserting into a QFileDialog.

    :param typeFilter: type filter for handled io types. For instanece, if typ='png', then
      a file filter list with only 'id.png' and 'class.png' will appear.
    """
    if isinstance(typeFilter, str):
      typeFilter = [typeFilter]
    fileFilters = []
    for typ, info in dict(**cls.handledIoTypes, **extraOpts).items():
      if any([t in typ for t in typeFilter]):
        fileFilters.append(f'{info} (*.{typ})')
    return ';;'.join(fileFilters)

  @classmethod
  def exportByFileType(cls, compDf: df, outFile: Union[str, Path], verifyIntegrity=True, **exportArgs):
    outFile = Path(outFile)
    outFn = cls._ioFnFromFileType(outFile, 'export')
    ret = outFn(compDf, outFile, **exportArgs)
    if verifyIntegrity and outFile.suffix[1:] in cls.roundTripIoTypes:
      matchingCols = np.setdiff1d(compDf.columns, [RTF.INST_ID,
                                                   RTF.SRC_IMG_FILENAME])
      loadedDf = cls.buildByFileType(outFile)
      dfCmp = loadedDf[matchingCols].equals(compDf[matchingCols])
      problemCells = defaultdict(list)

      if not dfCmp:
        dfA = loadedDf[matchingCols]
        dfB = compDf[matchingCols]
        for ii in range(len(dfA)):
          for jj in range(len(dfA.columns)):
            if not np.array_equal(dfA.iat[ii, jj], dfB.iat[ii,jj]):
              problemCells[compDf.at[dfB.index[ii], RTF.INST_ID]].append(str(matchingCols[jj]))
        # The only way to prevent "truth value of array is ambiguous" is cell-by-cell iteration
        problemMsg = [f'{idx}: {cols}' for idx, cols in problemCells.items()]
        problemMsg = '\n'.join(problemMsg)
        # Try to fix the problem with an iloc write
        warnLater('<b>Warning!</b> Saved components do not match current component'
             ' state. This can occur when pandas incorrectly caches some'
             ' table values. Problem cells (shown as [id]: [columns]):\n'
             + f'{problemMsg}\n'
               f'Please try manually altering these values before exporting again.', UserWarning)
    return ret

  @classmethod
  def buildByFileType(cls, inFile: Union[str, Path], imShape: Tuple[int]=None,
                      strColumns=False, **importArgs):
    buildFn = cls._ioFnFromFileType(inFile, 'buildFrom')
    outDf = buildFn(inFile, imShape=imShape, **importArgs)
    if strColumns:
      outDf.columns = list(map(str, outDf.columns))
    return outDf

  @classmethod
  def _ioFnFromFileType(cls, fpath: Union[str, Path],
                        buildOrExport=Literal['buildFrom', 'export']) -> Optional[Callable]:
    fpath = Path(fpath)
    fname = fpath.name
    cmpTypes = np.array(list(cls.handledIoTypes.keys()))
    typIdx = [fname.endswith(typ) for typ in cmpTypes]
    if not any(typIdx):
      raise IOError(f'Not sure how to handle file {fpath.stem}')
    fnNameSuffix = cmpTypes[typIdx][-1].title().replace('.', '')
    return getattr(cls, buildOrExport+fnNameSuffix, None)

  @staticmethod
  def _strToNpArray(array_string: str, **opts):
    # Adapted from https://stackoverflow.com/a/42756309/9463643
    array_string = ','.join(array_string.replace('[ ', '[').split())
    return np.array(literal_eval(array_string), **opts)

  @classmethod
  def _pandasCsvExport(cls, exportDf: pd.DataFrame, outFile: Union[str, Path]=None,
                       readOnly=True, **pdExportArgs):
    if outFile is None:
      return

    defaultExportParams = {
      'na_rep': 'NaN',
      'float_format': '{:0.10n}',
      'index': False,
    }
    outPath = Path(outFile)
    outPath.parent.mkdir(exist_ok=True, parents=True)

    defaultExportParams.update(pdExportArgs)
    with np.printoptions(threshold=sys.maxsize):
      exportDf.to_csv(outFile, index=False)
    if readOnly:
      outPath.chmod(S_IREAD|S_IRGRP|S_IROTH)

  # -----
  # Export options
  # -----
  @classmethod
  def exportCsv(cls, compDf: df, outFile: Union[str, Path]=None, readOnly=True, **pdExportArgs):
    """
    :param compDf: Dataframe to export
    :param outFile: Name of the output file location. If *None*, no file is created. However,
      the export object will still be created and returned.
    :param pdExportArgs: Dictionary of values passed to underlying pandas export function.
      These will overwrite the default options for :func:`exportToFile <ComponentMgr.exportToFile>`
    :param readOnly: Whether this export should be read-only
    :return: Export version of the component data.
    """
    # Make sure no rows are truncated
    col = None
    try:
      # TODO: Currently the additional options are causing errors. Find out why and fix
      #  them, since this may be useful if it can be modified
      # Format special columns appropriately
      # Since CSV export significantly modifies the df, make a copy before doing all these
      # operations
      exportDf = compDf.copy(deep=True)
      for col in exportDf:
        if not isinstance(col.value, str):
          exportDf[col] = _paramSerToStrSer(exportDf[col], col.value)
      cls._pandasCsvExport(exportDf, outFile, readOnly, **pdExportArgs)
    except Exception as ex:
      errMsg = f'Error on parsing column "{col.name}"\n'
      augmentException(ex, errMsg)
      raise
    return exportDf

  @classmethod
  def exportCompImgsDf(cls, compDf: df, outFile: Union[str, Path]=None,
                       imgDir: FilePath=None, margin=0, marginAsPct=False,
                       includeCols=('instId', 'img', 'labelMask', 'label', 'offset'),
                       lblField='Instance ID', allowOffset=True, **kwargs):
    """
    Creates a dataframe consisting of extracted images around each component
    :param compDf: Dataframe to export
    :param outFile: Where to save the result, if it should be saved. Caution -- this
      is currently a time-consuming process!
    :param imgDir: Where images corresponding to this dataframe are kept. Source image
      filenames are interpreted relative to this directory if they are not absolute.
    :param margin: How much padding to give around each component
    :param marginAsPct: Whether the margin should be a percentage of the component size or
      a raw pixel value.
    :param includeCols: Which columns to include in the export list
    :param lblField: See ComponentIO.exportLblPng. This label is provided in the output dataframe
      as well, if specified.
    :param allowOffset: See ComponentIO.exportLblPng. This ensures index labels start at 1
    :return: Dataframe with the following keys:
      - instId: The component's Instance ID
      - img: The (MxNxC) image corresponding to the component vertices, where MxN are
        the padded row sizes and C is the number of image channels
      - labelMask: Binary mask representing the component vertices
      - label: Field value of the component for the field specified by `lblField`
      - offset: Image (x,y) coordinate of the min component vertex.
    :param kwargs: Passed to ComponentIO.exportLblPng
    """
    _imgCache = {}
    if imgDir is None:
      imgDir = Path('.')
    else:
      imgDir = Path(imgDir)
    uniqueImgs = np.unique(compDf[RTF.SRC_IMG_FILENAME])
    dfGroupingsByImg = []
    for imgName in uniqueImgs:
      fullImgName = Path(imgName)
      if not fullImgName.is_absolute():
        fullImgName = imgDir / fullImgName
      if fullImgName not in _imgCache:
        _imgCache[fullImgName] = io.imread(fullImgName)
      dfGroupingsByImg.append(compDf[compDf[RTF.SRC_IMG_FILENAME] == imgName])

    useKeys = set(includeCols)
    outDf = {k: [] for k in useKeys}
    lblField = cls.tableData.fieldFromName(lblField)

    for miniDf, fullImgName in zip(dfGroupingsByImg, uniqueImgs):
      fullImgName = imgDir / fullImgName
      img = _imgCache[fullImgName]
      lblImg = cls.exportLblPng(miniDf, imShape=img.shape[:2], lblField=lblField,
                                allowOffset=allowOffset, **kwargs)

      for ii, (idx, row) in enumerate(miniDf.iterrows()):
        allVerts = row[RTF.VERTICES].stack()
        if marginAsPct:
          compImgSz = allVerts.max(0) - allVerts.min(0)
          marginToUse = (compImgSz*(margin/100)).astype(int)
        else:
          marginToUse = margin
        compImg, bounds = getCroppedImg(img, allVerts, marginToUse, coordsAsSlices=False)

        if 'img' in useKeys:
          outDf['img'].append(compImg)
        lbl = row[lblField]

        if 'label' in useKeys:
          outDf['label'].append(lbl)

        indexer = tuple(slice(*b) for b in bounds[:, ::-1].T)
        if lblImg.ndim > 2:
          indexer += (...,)
        mask = lblImg[indexer]
        if 'labelMask' in useKeys:
          # x-y to row-col, transpose to get min-max in axis 0 and row-col in axis 1
          outDf['labelMask'].append(mask)

        if 'instId' in useKeys:
          outDf['instId'].append(idx)

        if 'offset' in useKeys:
          outDf['offset'].append(bounds[0,:])
    outDf = pd.DataFrame(outDf)
    if outFile is not None:
      outDf.to_pickle(outFile)
    return outDf

  @classmethod
  def exportPkl(cls, compDf: df, outFile: Union[str, Path]=None, **exportArgs) -> (Any, str):
    """
    See the function signature for :func:`exportCsv <ComponentIO.exportCsv>`
    """
    # Since the write-out is a single operation there isn't an intermediate form to return
    pklDf = None
    if outFile is not None:
      pklDf = pickle.dumps(compDf)
      compDf.to_pickle(outFile)
    return pklDf

  @classmethod
  def exportLblPng(cls, compDf: df, outFile: FilePath=None, imShape: Tuple[int]=None,
                   lblField: Union[PrjParam, str]='Instance ID', bgColor=0, allowOffset=None,
                   rescaleOutput=False, returnLblMapping=False,
                   **kwargs):
    """
    :param compDf: Dataframe to export
    :param outFile: Filename to save, leave *None* to avoid saving to a file
    :param imShape: MxN shape of image containing these annotations
    :param lblField: Data field to use as an index label. E.g. "Class" will use the 'class'
      column, but any other column can be specified. The output ground truth masks
      will be colored according to this field.  See :meth:`PrjParam.toNumeric` for details.
      If `lblField` is *None*, the foreground mask will be boolean instead of integer-colored.
    :param bgColor: Color of the mask background. Must be an integer.
    :param allowOffset: Some label fields may have 0-based starting values. In cases
      where `bgColor` is also 0, an offset is required to prevent those field values
      from being considered background. `allowOffset` determines whether this change
      can be made to the data. See `PrjParam.toNumeric` for detailed description.
    :param rescaleOutput: For images designed for human use, it is helpful to have
      outputs rescaled to the entire intensity range. Otherwise, they usually end
      up looking pure black and it is difficult to see components in the image.
      When `rescaleOutput` is *True*, all numbers are scaled to the 'uint16' range.
    :param returnLblMapping: Whether to return a pd.Series matching original index values
      to their numeric counterparts
    :param kwargs:
    :return:
    """

    lblField = cls.tableData.fieldFromName(lblField)

    if lblField not in cls.tableData.allFields:
      raise IOError(f'Specified label field {lblField} does not exist in the table'
                       f' fields. Must be one of:\n'
                       f'{[f.name for f in cls.tableData.allFields]}')
    if bgColor < 0:
      raise IOError(f'Background color must be >= 0, was {bgColor}')

    labels = compDf[lblField]
    labels_numeric = lblField.toNumeric(labels, allowOffset)
    asBool = np.issubdtype(labels_numeric.dtype, np.bool_)

    if rescaleOutput:
      lowBound = 1 if allowOffset else 0
      upBound = np.iinfo(np.uint16).max
      labels_numeric = rescale_intensity(labels_numeric, out_range=(lowBound, upBound))

    if imShape is None:
      # Without any components the image is non-existant
      if len(compDf) == 0:
        raise ValueError('imShape cannot be *None* if no components are present')
      vertMax = ComplexXYVertices.stackedMax(compDf[RTF.VERTICES])
      imShape = tuple(vertMax[::-1] + 1)
    maskType = 'uint16' if np.min(bgColor) >= 0 else 'int32'
    outMask = np.full(imShape[:2], bgColor, dtype=maskType)
    for fillClr, (_, comp) in zip(labels_numeric, compDf.iterrows()):
      verts: ComplexXYVertices = comp[RTF.VERTICES]
      outMask = verts.toMask(outMask, int(fillClr), False, False)
    if asBool:
      outMask = outMask > 0

    if outFile is not None:
      cvImsave_rgb(outFile, outMask.astype('uint16'))
    if returnLblMapping:
      mapping = pd.Series(data=labels_numeric, index=labels, name=lblField)
      return outMask, mapping
    return outMask

  @classmethod
  def exportClassPng(cls, compDf: df, outFile: FilePath = None, imShape: Tuple[int]=None, **kwargs):
    # Create label to output mapping
    return cls.exportLblPng(compDf, outFile, imShape, 'Class', **kwargs)

  @classmethod
  def exportIdPng(cls, compDf: df, outFile: FilePath=None,
                  imShape: Tuple[int]=None, **kwargs):
    """
    Creates a 2D grayscale image where each component is colored with its isntance ID + 1.
    *Note* Since Id 0 would end up not coloring the mask, all IDs must be offest by 1.
    :param compDf: Dataframe to export
    :param imShape: The size of this output image
    :param outFile: Where to save the output. If *None*, no export is created.
    :return:
    """
    return cls.exportLblPng(compDf, outFile, imShape, **kwargs, allowOffset=True)

  @classmethod
  def exportCompImgsZip(cls, compDf: df,
                            outDir:FilePath='s3a-export',
                            resizeShape: Tuple[int, int]=None,
                            **kwargs):
    """
    From a component dataframe, creates output directories for component images and masks.
    This is useful for many neural networks etc. to read individual component images.

    :param compDf: Dataframe to export
    :param outDir: Where to make the output directories. If `None`, defaults to current
      directory>compimgs_<margin>_margin
    :param resizeShape: If provided, it is the shape that all images will be resized to before
      being saved. This is useful for neural networks with a fixed input size which forces all
      inputs to be e.g. 100x100 pixels.
    :param kwargs: Passed directly to :meth:`ComponentIO.exportCompImgsDf`
    """
    outDir = Path(outDir)
    dataDir = outDir/'data'
    labelsDir= outDir/'labels'
    dataDir.mkdir(exist_ok=True, parents=True)
    labelsDir.mkdir(exist_ok=True, parents=True)
    saveFn = lambda fname, img: cvImsave_rgb(fname, img)
    if resizeShape is not None:
      saveFn = lambda fname, img: cvImsave_rgb(fname, resize_pad(img, resizeShape))

    extractedImgs = cls.exportCompImgsDf(compDf, None, **kwargs)
    for idx, row in extractedImgs.iterrows():
      saveName = f'{row.instId}.png'
      saveFn(dataDir/saveName, row.img)
      saveFn(labelsDir/saveName, row.labelMask)

  # -----
  # Import options
  # -----

  @classmethod
  def convert(cls, fromData: FilePathOrDf, toFile: FilePath, doExport=True, importArgs: dict=None,
              exportArgs: dict=None):
    if importArgs is None:
      importArgs = {}
    if exportArgs is None:
      exportArgs = {}
    if not isinstance(fromData, df):
      fromData = cls.buildByFileType(fromData, **importArgs)
    exportFn = cls._ioFnFromFileType(toFile, 'export')
    if not doExport:
      toFile = None
    return exportFn(fromData, toFile, **exportArgs)

  @classmethod
  def buildFromCsv(cls, inFileOrDf: FilePathOrDf, imShape: Tuple=None,
                   reindex=False) -> df:
    """
    Deserializes data from a csv file to create a Component :class:`DataFrame`.
    The input .csv should be the same format as one exported by
    :func:`csvImport <ComponentMgr.csvImport>`.

    :param imShape: If included, this ensures all imported components lie within imSize
           boundaries. If any components do not, an error is thrown since this is
           indicative of components that actually came from a different reference image.
    :param inFileOrDf: Name of file to import, or dataframe if it was already read from this
      file type. Useful if several csv's were concatenated into one dataframe and *that* is
      being imported.
    :param reindex: Whether to disregard the index of the incoming dataframe or file.
      This is useful when *inFileOrDf* is actually a conacatenated df of multiple files, and
      the index doesn't need to be retained.
    :return: Tuple: DF that will be exported if successful extraction
    """
    field = PrjParam('None', None)
    try:
      if isinstance(inFileOrDf, df):
        csvDf = inFileOrDf
      else:
        csvDf = pd.read_csv(inFileOrDf, keep_default_na=False)
      if reindex:
        csvDf[RTF.INST_ID.name] = np.arange(len(csvDf), dtype=int)
        csvDf = csvDf.set_index(RTF.INST_ID.name, drop=False)
      # Decouple index from instance ID until after transfer from csvDf is complete
      # This was causing very strange behavior without reset_index()...
      outDf = cls.tableData.makeCompDf(len(csvDf)).reset_index(drop=True)
      # Objects in the original frame are represented as strings, so try to convert these
      # as needed
      for field in cls.tableData.allFields:
        if field.name in csvDf:
          matchingCol = csvDf[field.name]
          # 'Object' type results in false positives
          if matchingCol.dtype != object and type(field.value) == matchingCol.dtype:
            outDf[field] = csvDf[field.name]
          else:
            # Parsing functions only know how to convert from strings to themselves.
            # So, assume the exting types can first convert themselves to strings
            with np.printoptions(threshold=sys.maxsize):
              matchingCol = matchingCol.apply(str)
            outDf[field] = _strSerToParamSer(matchingCol, field.value)
      outDf = outDf.set_index(RTF.INST_ID, drop=False)

      cls.checkVertBounds(outDf[RTF.VERTICES], imShape)
    except Exception as ex:
      # Rethrow exception with insight about column number
      # Procedure copied from https://stackoverflow.com/a/6062677/9463643
      errMsg = f'Error importing column "{field.name}":\n'
      augmentException(ex, errMsg)
      raise
    # TODO: Apply this function to individual rows instead of the whole dataframe. This will allow malformed
    #  rows to gracefully fall off the dataframe with some sort of warning message
    return outDf

  @classmethod
  def buildFromCompimgsDf(cls, inFile: FilePath, imShape: Tuple=None,
                          lblField='Class', **importArgs):
    lblField = cls.tableData.fieldFromName(lblField)
    inDf = pd.read_pickle(inFile)
    outDf = cls.tableData.makeCompDf(len(inDf))
    outDf[RTF.INST_ID] = inDf['instId']
    allVerts = []

    for idx, row in inDf.iterrows():
      mask = cls._strToNpArray(row.labelMask, dtype=bool)
      verts = ComplexXYVertices.fromBwMask(mask)
      offset = cls._strToNpArray(row.offset)
      for v in verts: v += offset
      allVerts.append(verts)
    outDf[RTF.VERTICES] = allVerts
    outDf[lblField] = inDf[lblField]
    cls.checkVertBounds(outDf[RTF.VERTICES], imShape)
    return outDf

  @classmethod
  def buildFromPkl(cls, inFile: FilePath, imShape: Tuple=None, **importArgs) -> df:
    """
    See docstring for :func:`self.buildFromCsv`
    """
    pklDf = pd.read_pickle(inFile)
    cls.checkVertBounds(pklDf[RTF.VERTICES], imShape)
    templateDf = cls.tableData.makeCompDf(len(pklDf)).set_index(pklDf.index)
    templateDf.update(pklDf)
    return templateDf

  @classmethod
  def buildFromLblPng(cls, inFileOrImg: Union[FilePath, GrayImg],
                      labelMapping: pd.Series=None,
                      useDistinctRegions=False,
                      **importArgs) -> df:
    if isinstance(inFileOrImg, GrayImg):
      labelImg = inFileOrImg
    else:
      labelImg = io.imread(inFileOrImg, as_gray=True)
    allVerts = []
    lblField_out = []
    for origVal, numericLbl in labelMapping.iteritems():
      verts = ComplexXYVertices.fromBwMask(labelImg == numericLbl)
      if useDistinctRegions:
        allVerts.extend(verts)
        orig = np.tile(origVal, len(verts))
      else:
        allVerts.append(verts)
        orig = [origVal]
      lblField_out.extend(orig)
    outDf = cls.tableData.makeCompDf(len(allVerts))
    outDf[labelMapping.name] = lblField_out
    outDf[RTF.VERTICES] = allVerts
    return outDf


  @classmethod
  def buildFromIdPng(cls, inFileOrImg: Union[FilePath, GrayImg], imShape: Tuple=None, **importArgs) -> df:
    if isinstance(inFileOrImg, GrayImg):
      labelImg = inFileOrImg
    else:
      labelImg = io.imread(inFileOrImg, as_gray=True)
    outDf = cls._idImgToDf(labelImg)
    cls.checkVertBounds(outDf[RTF.VERTICES], imShape)
    return outDf

  @classmethod
  def buildFromClassPng(cls, inFileOrImg: Union[FilePath, GrayImg], imShape: Tuple=None, **importArgs) -> df:
    if isinstance(inFileOrImg, GrayImg):
      clsImg = inFileOrImg
    else:
      clsImg = io.imread(inFileOrImg)
    clsParam = cls.tableData.fieldFromName('Class')
    clsArray = np.array(clsParam.opts['limits'])
    idImg = measure.label(clsImg)
    outDf = cls.buildFromIdPng(idImg, imShape)
    outClasses = []
    for curId in outDf[RTF.INST_ID]+1:
      # All ID pixels should be the same class, so any representative will do
      curCls = clsImg[idImg == curId][0]
      outClasses.append(clsArray[curCls-1])
    outDf[cls.tableData.fieldFromName('Class')] = outClasses
    outDf.reset_index(inplace=True, drop=True)
    outDf[RTF.INST_ID] = outDf.index
    return outDf

  @staticmethod
  def checkVertBounds(vertSer: pd.Series, imShape: tuple):
    """
    Checks whether any vertices in the imported dataframe extend past image dimensions. This is an indicator
    they came from the wrong import file.

    :param vertSer: Vertices from incoming component dataframe
    :param imShape: Shape of the main image these vertices are drawn on
    :return: Raises error if offending vertices are present, since this is an indication the component file
      was from a different image
    """
    if imShape is None or len(vertSer) == 0:
      # Nothing we can do if no shape is given
      return
    # Image shape from row-col -> x-y
    imShape = np.array(imShape[1::-1])[None, :]
    # Remove components whose vertices go over any image edges
    vertMaxs = [verts.stack().max(0) for verts in vertSer if len(verts) > 0]
    vertMaxs = np.vstack(vertMaxs)
    offendingIds = np.nonzero(np.any(vertMaxs >= imShape, axis=1))[0]
    if len(offendingIds) > 0:
      warnLater(f'Vertices on some components extend beyond image dimensions. '
           f'Perhaps this export came from a different image?\n'
           f'Offending IDs: {offendingIds}', UserWarning)

  @classmethod
  def _idImgToDf(cls, idImg: GrayImg):
    # Skip 0 since it's indicative of background
    regionIds = np.unique(idImg)
    regionIds = regionIds[regionIds != 0]
    allVerts = []
    for curId in regionIds:
      verts = ComplexXYVertices.fromBwMask(idImg == curId)
      allVerts.append(verts)
    outDf = cls.tableData.makeCompDf(regionIds.size)
    # Subtract 1 since instance ids are 0-indexed
    outDf[RTF.INST_ID] = regionIds-1
    outDf[RTF.VERTICES] = allVerts
    return outDf


def ALL_PLUGINS():
  from .misc import MainImagePlugin, CompTablePlugin, RandomToolsPlugin, EditPlugin, HelpPlugin, MultiPredictionsPlugin
  from .file import FilePlugin
  from .tablefield import VerticesPlugin

  return [VerticesPlugin, EditPlugin, MultiPredictionsPlugin, MainImagePlugin, CompTablePlugin, RandomToolsPlugin, HelpPlugin]


from functools import wraps
from typing import Union, Collection, Callable, Any, Sequence, List

import numpy as np
import pandas as pd
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

from s3a import PRJ_SINGLETON
from s3a.constants import REQD_TBL_FIELDS as RTF, PRJ_CONSTS as CNST
from s3a.controls.drawctrl import RoiCollection
from s3a.generalutils import getCroppedImg, coerceDfTypes
from s3a.structures import XYVertices
from utilitys import ParamEditor, PrjParam, RunOpts, fns, EditorPropsMixin
from utilitys.widgets import ButtonCollection, ImageViewer, EasyWidget
from .clickables import RightPanViewBox
from .regions import RegionCopierPlot
from .rois import SHAPE_ROI_MAPPING

__all__ = ['MainImage']

Signal = QtCore.Signal
QCursor = QtGui.QCursor

DrawActFn = Union[Callable[[XYVertices, PrjParam], Any], Callable[[XYVertices], Any]]

class MainImage(EditorPropsMixin, ImageViewer):
  sigShapeFinished = Signal(object, object)
  """
  (XYVertices, PrjParam) emitted when a shape is finished
  - XYVerts from roi, re-thrown from self.shapeCollection
  - Current draw action
  """

  sigUpdatedFocusedComp = Signal(object)
  """pd.Series, newly focused component"""

  minCompSize: int

  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()

    cls.minCompSize = PRJ_SINGLETON.generalProps.registerProp(CNST.PROP_MIN_COMP_SZ)

  def __init__(self, parent=None, drawShapes: Collection[PrjParam]=None,
               toolbar: QtWidgets.QToolBar=None,
               **kargs):
    super().__init__(parent, viewBox=RightPanViewBox(), **kargs)

    if drawShapes is None:
      drawShapes = SHAPE_ROI_MAPPING.keys()

    self._initGrid()
    PRJ_SINGLETON.colorScheme.registerFunc(self.updateGridScheme, runOpts=RunOpts.ON_CHANGED)
    self.lastClickPos = QtCore.QPoint()
    self.toolbar = toolbar

    self.toolsEditor.registerFunc(self.resetZoom, btnOpts=CNST.TOOL_RESET_ZOOM)

    # -----
    # FOCUSED COMPONENT INFORMATION
    # -----
    self.compSer: pd.Series = PRJ_SINGLETON.tableData.makeCompSer()
    self._focusedTools: List[ParamEditor] = []
    """
    List of all toolsEditor that allow actions to be performed on the currently focused components
    """
    # -----
    # DRAWING OPTIONS
    # -----
    self.regionCopier = RegionCopierPlot(self)

    self.drawAction: PrjParam = CNST.DRAW_ACT_PAN
    self.shapeCollection = RoiCollection(drawShapes, self)
    self.shapeCollection.sigShapeFinished.connect(
      lambda roiVerts: self.sigShapeFinished.emit(roiVerts, self.drawAction)
    )

    self.drawShapeGrp = ButtonCollection(self, 'Shapes', drawShapes, self.shapeAssignment,
                                         namePath=(self.__groupingName__,),
                                         checkable=True)
    self.drawActGrp = ButtonCollection(self, 'Actions')

    # Make sure panning is allowed before creating draw widget
    self.registerDrawAction(CNST.DRAW_ACT_PAN, lambda *args: self.actionAssignment(CNST.DRAW_ACT_PAN))

    # Initialize draw shape/action buttons
    self.drawActGrp.callFuncByParam(self.drawAction)
    self.drawShapeGrp.callFuncByParam(self.shapeCollection.curShapeParam)

    self.toolsGrp = None
    if toolbar is not None:
      toolbar.addWidget(self.drawShapeGrp)
      toolbar.addWidget(self.drawActGrp)
      self.toolsGrp = self.addTools(self.toolsEditor)

  def resetZoom(self):
    """
    Reimplement viewbox zooming since padding required for scatterplot with
    dynamic uncached shapes means the viewbox overestimates required rect
    """
    if self.image is None:
      return
    imShape = self.image.shape
    vb: RightPanViewBox = self.getViewBox()
    vb.setRange(xRange=(0, imShape[1]), yRange=(0, imShape[0]))

  @property
  def compSer_asFrame(self):
      return coerceDfTypes(fns.serAsFrame(self.compSer))

  def shapeAssignment(self, newShapeParam: PrjParam):
    self.shapeCollection.curShapeParam = newShapeParam
    if self.regionCopier.active:
      self.regionCopier.erase()

  def actionAssignment(self, newActionParam: PrjParam):
    self.drawAction = newActionParam
    if self.regionCopier.active:
      self.regionCopier.erase()

  def maybeBuildRoi(self, ev: QtGui.QMouseEvent):
    ev.ignore()
    if (QtCore.Qt.LeftButton not in [ev.buttons(), ev.button()]
        or self.drawAction == CNST.DRAW_ACT_PAN
        or self.regionCopier.active):
      return False
    finished = self.shapeCollection.buildRoi(ev, self.imgItem)
    if not finished:
      ev.accept()
    return finished

  def switchBtnMode(self, newMode: PrjParam):
    # EAFP
    try:
      self.drawActGrp.callFuncByParam(newMode)
    except KeyError:
      self.drawShapeGrp.callFuncByParam(newMode)

  def mousePressEvent(self, ev: QtGui.QMouseEvent):
    self.maybeBuildRoi(ev)
    if not ev.isAccepted():
      super().mousePressEvent(ev)
    self.lastClickPos = ev.pos()

  def mouseDoubleClickEvent(self, ev: QtGui.QMouseEvent):
    if self.regionCopier.active:
      self.regionCopier.sigCopyStopped.emit()
      ev.accept()
      self.shapeCollection.addLock(self)
    self.maybeBuildRoi(ev)
    if not ev.isAccepted():
      super().mouseDoubleClickEvent(ev)

  def mouseMoveEvent(self, ev: QtGui.QMouseEvent):
    """
    Mouse move behavior is contingent on which shape is currently selected,
    unless we are panning
    """
    self.maybeBuildRoi(ev)
    if not ev.isAccepted():
      super().mouseMoveEvent(ev)

  def mouseReleaseEvent(self, ev: QtGui.QMouseEvent):
    # Typical reaction is to right-click to cancel an roi
    if self.image is not None and QtCore.Qt.RightButton not in [ev.button(), ev.buttons()]:
      self.maybeBuildRoi(ev)

      # Special case: Panning
      if (self.lastClickPos == ev.pos()
          and self.drawAction == CNST.DRAW_ACT_PAN
          and not self.regionCopier.active):
        pos = self.imgItem.mapFromScene(ev.pos())
        xx, yy, = pos.x(), pos.y()
        # Simulate a click-wide boundary selection so points can be registered in pan mode
        pt = XYVertices([[xx, yy]], dtype=float)
        self.shapeCollection.sigShapeFinished.emit(pt)
    super().mouseReleaseEvent(ev)
    self.shapeCollection.removeLock(self)

  def clearCurRoi(self):
    """Clears the current ROI shape"""
    self.shapeCollection.clearAllRois()
    self.regionCopier.erase()

  def _widgetContainerChildren(self):
    return [self.drawActGrp, self.drawShapeGrp, self]

  def _initGrid(self):
    pi: pg.PlotItem = self.plotItem
    pi.showGrid(alpha=1.0)
    axs = [pi.getAxis(ax) for ax in ['top', 'bottom', 'left', 'right']]
    for ax in axs: # type: pg.AxisItem
      ax.setZValue(1e9)
      # ax.setTickSpacing(5, 1)
      for evFn in 'mouseMoveEvent', 'mousePressEvent', 'mouseReleaseEvent', \
                  'mouseDoubleClickEvent', 'mouseDragEvent', 'wheelEvent':
        newEvFn = lambda ev: ev.ignore()
        setattr(ax, evFn, newEvFn)

  def updateGridScheme(self, showGrid=False, gridWidth=1, gridColor='fff'):
    """
    :param showGrid:
    :param gridWidth:
    :param gridColor:
      pType: color
    """
    pi: pg.PlotItem = self.plotItem
    pi.showGrid(showGrid, showGrid)
    axs = [pi.getAxis(ax) for ax in ['top', 'bottom', 'left', 'right']]
    newPen = pg.mkPen(width=gridWidth, color=gridColor)
    for ax in axs:
      ax.setPen(newPen)

  def registerDrawAction(self, actParams: Union[PrjParam, Sequence[PrjParam]], func: DrawActFn,
                         **registerOpts):
    """
    Adds specified action(s) to the list of allowable roi actions if any do not already
    exist. `func` is only triggered if a shape was finished and the current action matches
    any of the specified `actParams`

    :param actParams: Single :py:class:`~s3a.structures.PrjParam` or multiple PrjParams that are allowed
      to trigger this funciton. If empty, triggers on every parameter
    :param func: Function to trigger when a shape is completed during the requested actions.
      If only one parameter is registered to this function, it is expected to only take
      roiVerts. If multiple or none are provided, it is expected to take roiVerts and the current draw action
    :param registerOpts: Extra arguments for button registration
    """
    if isinstance(actParams, PrjParam):
      actParams = [actParams]

    @wraps(func)
    def wrapper(roiVerts: XYVertices, param: PrjParam):
      if param in actParams:
        if len(actParams) > 1:
          func(roiVerts, param)
        else:
          func(roiVerts)
    if len(actParams) == 0:
      self.sigShapeFinished.connect(func)
    else:
      self.sigShapeFinished.connect(wrapper)
    for actParam in actParams:
      self.drawActGrp.create_addBtn(actParam, self.actionAssignment, checkable=True,
                                    namePath=(self.__groupingName__,), **registerOpts)

  def focusedCompImage(self, margin: int=0):
    return getCroppedImg(self.image, self.compSer[RTF.VERTICES].stack(), margin, returnSlices=False)

  def viewboxSquare(self, margin=0):
    vbRange = np.array(self.getViewBox().viewRange())
    span = np.diff(vbRange).flatten()
    center = vbRange[:,0]+span/2
    minSpan = np.min(span) + margin
    offset = center - minSpan/2
    return minSpan*np.array([[0,0], [0,1], [1,1], [1,0]]) + offset

  def addTools(self, toolsEditor: ParamEditor):
    if toolsEditor in self._focusedTools:
      return
    toolsEditor.actionsMenuFromProcs(outerMenu=self.menu, parent=self, nest=True)
    # self.toolsGrp.clear()
    # self.toolsGrp.fromToolsEditors(self._focusedTools, checkable=False, ownerClctn=self.toolsGrp)
    retClctn = None
    if self.toolbar is not None:
      retClctn = ButtonCollection.fromToolsEditors([toolsEditor], title=toolsEditor.name, copy=False)
      toolsEditor.params.sigChildAdded.connect(
        lambda _param, child, _idx: retClctn.addByParam(child, copy=False)
      )
      self.toolbar.addWidget(retClctn)
    self.getViewBox().menu = self.menu
    return retClctn

  @PRJ_SINGLETON.actionStack.undoable('Update Focused Component')
  def updateFocusedComp(self, newComp: pd.Series=None):
    """
    Updates focused image and component from provided information. Useful for creating
    a 'zoomed-in' view that allows much faster processing than applying image processing
    algorithms to the entire image each iteration.
    :param newComp: New component to edit using various plugins (See :class:`TableFieldPlugin`)
    """
    oldComp = self.compSer
    mainImg = self.image
    if newComp is None or mainImg is None:
      newComp = PRJ_SINGLETON.tableData.makeCompSer()
    else:
      # Since values INSIDE the dataframe are reset instead of modified, there is no
      # need to go through the trouble of deep copying
      newComp = newComp.copy(deep=False)

    if mainImg is None:
      self.imgItem.clear()
      self.shapeCollection.clearAllRois()
    self.compSer = newComp

    self.sigUpdatedFocusedComp.emit(newComp)
    yield
    self.updateFocusedComp(oldComp)


from typing import Union, Sequence
from warnings import warn

import numpy as np
import pyqtgraph as pg
from pandas import DataFrame as df
from pyqtgraph.Qt import QtCore, QtGui, QtSvg, QtWidgets
from utilitys import EditorPropsMixin, PrjParam, RunOpts
from utilitys.fns import warnLater

from s3a import PRJ_SINGLETON
from s3a.constants import PRJ_CONSTS, REQD_TBL_FIELDS, PRJ_ENUMS
from s3a.models.tablemodel import ComponentMgr
from s3a.structures import OneDArr
from s3a.structures import XYVertices, ComplexXYVertices
from s3a.views import tableview
from s3a.views.imageareas import MainImage
from s3a.views.regions import MultiRegionPlot

__all__ = ['CompSortFilter', 'CompDisplayFilter']

Signal = QtCore.Signal
TBL_FIELDS = PRJ_SINGLETON.tableData.allFields
QISM = QtCore.QItemSelectionModel

class CompSortFilter(QtCore.QSortFilterProxyModel):
  colTitles = [f.name for f in TBL_FIELDS]
  def __init__(self, compMgr: ComponentMgr, parent=None):
    super().__init__(parent)
    self.setSourceModel(compMgr)
    # TODO: Move code for filtering into the proxy too. It will be more efficient and
    #  easier to generalize than the current solution in CompDisplayFilter.


  def sort(self, column: int, order: QtCore.Qt.SortOrder=...) -> None:
    # Do nothing if the user is trying to sort by vertices, since the intention of
    # sorting numpy arrays is somewhat ambiguous
    noSortCols = []
    for ii, col in enumerate(PRJ_SINGLETON.tableData.allFields):
      if isinstance(col.value, (list, np.ndarray)):
        noSortCols.append(ii)
    if column in noSortCols:
      return
    else:
      super().sort(column, order)

  def lessThan(self, left: QtCore.QModelIndex, right: QtCore.QModelIndex) -> bool:
    # First, attempt to compare the object data
    leftObj = left.data(QtCore.Qt.EditRole)
    rightObj = right.data(QtCore.Qt.EditRole)
    try:
      return np.all(leftObj < rightObj)
    except (ValueError, TypeError):
      # If that doesn't work, default to stringified comparison
      return str(leftObj) < str(rightObj)

class CompDisplayFilter(EditorPropsMixin, QtCore.QObject):
  sigCompsSelected = Signal(object)

  __groupingName__ = 'Main Image'

  @classmethod
  def __initEditorParams__(cls):
    cls.pltClickBehav: str = PRJ_SINGLETON.generalProps.registerProp(
      PRJ_CONSTS.PROP_COMP_SEL_BHV)

  def __init__(self, compMgr: ComponentMgr, mainImg: MainImage,
               compTbl: tableview.CompTableView, parent=None):
    super().__init__(parent)
    filterEditor = PRJ_SINGLETON.filter
    self._mainImgArea = mainImg
    self._filter = filterEditor
    self._compTbl = compTbl
    self._compMgr = compMgr
    self.regionPlot = MultiRegionPlot()
    self.displayedIds = np.array([], dtype=int)
    self.selectedIds = np.array([], dtype=int)
    self.labelCol = REQD_TBL_FIELDS.INST_ID
    self.updateLabelCol()

    self.regionCopier = mainImg.regionCopier

    with PRJ_SINGLETON.colorScheme.setBaseRegisterPath(self.regionPlot.__groupingName__):
      proc, argsParam = PRJ_SINGLETON.colorScheme.registerFunc(
        self.updateLabelCol, runOpts=RunOpts.ON_CHANGED, returnParam=True, nest=False)
    def updateLblList():
      fields = PRJ_SINGLETON.tableData.allFields
      # TODO: Filter out non-viable field types
      argsParam.child('labelCol').setLimits([f.name for f in fields])
    PRJ_SINGLETON.tableData.sigCfgUpdated.connect(updateLblList)

    # Attach to UI signals
    def _maybeRedraw():
      """
      Since an updated filter can also result from refreshed table fields, make sure not to update in that case
      (otherwise errors may occur from missing classes, etc.)
      """
      if np.array_equal(PRJ_SINGLETON.tableData.allFields, self._compMgr.compDf.columns):
        self.redrawComps()
    self._filter.sigChangesApplied.connect(_maybeRedraw)

    self.regionCopier.sigCopyStarted.connect(lambda *args: self.activateRegionCopier())
    self.regionCopier.sigCopyStopped.connect(lambda *args: self.finishRegionCopier())

    compMgr.sigCompsChanged.connect(self.redrawComps)
    compMgr.sigFieldsChanged.connect(lambda: self._reflectFieldsChanged())
    compTbl.sigSelectionChanged.connect(self._reflectTableSelectionChange)

    mainImg.addItem(self.regionPlot)
    mainImg.addItem(self.regionCopier)

    # self.filterableCols = self.findFilterableCols()

  def updateLabelCol(self, labelCol=REQD_TBL_FIELDS.INST_ID.name):
    """
    Changes the data column used to label (color) the region plot data
    :param labelCol:
      helpText: New column to use
      title: Labeling Column
      pType: list
    """
    self.labelCol = PRJ_SINGLETON.tableData.fieldFromName(labelCol)
    newLblData = self.labelCol.toNumeric(self._compMgr.compDf[self.labelCol], rescale=True)

    self.regionPlot.regionData[PRJ_ENUMS.FIELD_LABEL] = newLblData
    self.regionPlot.updateColors()

  def redrawComps(self, idLists=None):
    # Following mix of cases are possible:
    # Components: DELETED, UNCHANGED, CHANGED, NEW
    # New is different from changed since id plot already exists (unhide vs. create)
    # Plots: DRAWN, UNDRAWN
    # Note that hiding the ID is chosen instead of deleting, since that is a costly graphics
    # operation
    compDf = self._compMgr.compDf

    # Update and add changed/new components
    # TODO: Find out why this isn't working. For now, just reset the whole comp list
    #  each time components are changed, since the overhead isn't too terrible.
    # changedIds = np.concatenate((idLists['added'], idLists['changed']))
    # self._regionPlots[changedIds, regCols] = compDf.loc[changedIds, compCols]

    # Hide all ids and table rows, since they will be reshown as needed after display filtering
    for rowIdx in range(len(compDf)):
      self._compTbl.hideRow(rowIdx)

    # Component deleted: Nothing to do, since only displayed IDs will remain in the
    # region manager anyway
    #idsToRm = idLists['deleted']

    # Update filter list: hide/unhide ids and verts as needed.
    self._updateDisplayedIds()
    # Remove all IDs that aren't displayed
    # FIXME: This isn't working correctly at the moment
    # self._regionPlots.drop(np.setdiff1d(self._regionPlots.data.index, self._displayedIds))
    self.regionPlot.resetRegionList(compDf.loc[self.displayedIds], lblField=self.labelCol)
    # noinspection PyTypeChecker
    # self._reflectTableSelectionChange(np.intersect1d(self.displayedIds, self.selectedIds))

    tblIdsToShow = np.isin(compDf.index, self.displayedIds).nonzero()[0]
    model = self._compTbl.model()
    for rowId in tblIdsToShow:
      xpondingIdx = model.mapFromSource(self._compMgr.index(rowId,0)).row()
      self._compTbl.showRow(xpondingIdx)

  @PRJ_SINGLETON.actionStack.undoable('Split Components', asGroup=True)
  def splitSelectedComps(self):
    """Makes a separate component for each distinct boundary of all selected components"""
    selection = self._compTbl.ids_rows_colsFromSelection(excludeNoEditCols=False,
                                                            warnNoneSelection=False)
    if len(selection) == 0:
      return
    changes = self._compMgr.splitCompVertsById(np.unique(selection[:,0]))
    self.selectRowsById(changes['added'], QISM.ClearAndSelect)

  @PRJ_SINGLETON.actionStack.undoable('Merge Components', asGroup=True)
  def mergeSelectedComps(self, keepId=-1):
    """
    Merges the selected components into one, keeping all properties of the first in the selection
    :param keepId: If specified and >0, this is the ID whose peripheral data will be retained
      during the merge. Otherwise, the first selected component is used as the keep ID.
    """
    selection = self._compTbl.ids_rows_colsFromSelection(excludeNoEditCols=False,
                                                         warnNoneSelection=False)

    if len(selection) < 2:
      # Nothing to do
      return
    if keepId < 0:
      keepId = selection[0,0]
    try:
      self._compMgr.mergeCompVertsById(np.unique(selection[:,0]), keepId)
    except UserWarning:
      # No merge was performed, don't alter the table selection
      raise
    else:
      self.selectRowsById(np.array([keepId]), QISM.ClearAndSelect)


  def removeSelectedCompOverlap(self):
    """
    Makes sure all specified components have no overlap. Preference is given
    in order of the selection, i.e. the last selected component in the list
    is guaranteed to keep its full shape.
    """
    if self.selectedIds.size == 0:
      return
    self._compMgr.removeOverlapById(self.selectedIds)

  def _reflectFieldsChanged(self):
    self.redrawComps()

  def _reflectTableSelectionChange(self, selectedIds: OneDArr):
    self.selectedIds = selectedIds
    self.regionPlot.selectById(selectedIds)
    selectedComps = self._compMgr.compDf.loc[selectedIds]
    self.sigCompsSelected.emit(selectedComps)

  def scaleViewboxToSelectedIds(self, selectedIds: OneDArr=None, padding: int=None):
    """
    Rescales the main image viewbox to encompass the selection

    :param selectedIds: Ids to scale to. If *None*, this is the current selection
    :param padding: Padding around the selection. If *None*, defaults to
      pyqtgraph padding behavior
    """
    if selectedIds is None:
      selectedIds = self.selectedIds
    if len(selectedIds) == 0: return
    # Calculate how big the viewbox needs to be
    selectedVerts = self._compMgr.compDf.loc[selectedIds, REQD_TBL_FIELDS.VERTICES]
    allVerts = np.vstack([v.stack() for v in selectedVerts])
    mins = allVerts.min(0)
    maxs = allVerts.max(0)
    if padding is not None:
      mins -= padding//2
      maxs += padding//2
    vb: pg.ViewBox = self._mainImgArea.getViewBox()
    viewRect = QtCore.QRectF(*mins, *(maxs - mins))
    vb.setRange(viewRect, padding=padding)

  def selectRowsById(self, ids: Sequence[int],
                     selectionMode=QISM.Rows|QISM.ClearAndSelect,
                     onlyEditableRetList=True):
    selectionModel = self._compTbl.selectionModel()
    sortModel = self._compTbl.model()
    isFirst = True
    shouldScroll = len(ids) > 0
    selectionList = QtCore.QItemSelection()
    retLists = [] # See tableview ids_rows_colsFromSelection
    if onlyEditableRetList:
      selectedCols = self._compMgr.editColIdxs
    else:
      selectedCols = np.arange(len(self._compMgr.colTitles))
    for curId in ids:
      idRow = np.nonzero(self._compMgr.compDf.index == curId)[0][0]
      # Map this ID to its sorted position in the list
      idxForId = sortModel.mapFromSource(self._compMgr.index(idRow, 0))
      selectionList.select(idxForId, idxForId)
      if isFirst and shouldScroll:
        self._compTbl.scrollTo(idxForId, self._compTbl.PositionAtCenter)
        isFirst = False
      tblRow = idxForId.row()
      retLists.extend([[curId, tblRow, col] for col in selectedCols])
    # noinspection PyTypeChecker
    selectionModel.select(selectionList, selectionMode)
    return np.array(retLists)
    # if int(selectionMode & QISM.ClearAndSelect) > 0:
    #   self.selectedIds = ids
    # else: # Add to selection without clearing old selection
    #   self.selectedIds = np.concatenate([self.selectedIds, ids])


  def reflectSelectionBoundsMade(self, selection: Union[OneDArr, XYVertices]):
    """
    :param selection: bounding box of user selection: [xmin ymin; xmax ymax]
    """
    # If min and max are the same, just check for points at mouse position
    if selection.size == 0: return
    if len(selection) == 1 or np.abs(selection[0] - selection[1]).sum() < 0.01:
      qtPoint = QtCore.QPointF(*selection[0])
      selectedSpots = self.regionPlot.pointsAt(qtPoint, self.pltClickBehav=='Boundary Only')
      selectedIds = [spot.data() for spot in selectedSpots]
    else:
      selectedIds = self.regionPlot.boundsWithin(selection)
      selectedIds = np.unique(selectedIds)

    # -----
    # Obtain table idxs corresponding to ids so rows can be highlighted
    # -----
    # Add to current selection depending on modifiers
    mode = QISM.Rows
    if QtGui.QGuiApplication.keyboardModifiers() == QtCore.Qt.ControlModifier:
      mode |= QISM.Select
    else:
      mode |= QISM.ClearAndSelect
    if not self.regionCopier.active:
      self.selectRowsById(selectedIds, mode)
    # TODO: Better management of widget focus here

  def _updateDisplayedIds(self):
    curComps = self._filter.filterCompDf(self._compMgr.compDf.copy())
    # Give self the id list of surviving comps
    self.displayedIds = curComps[REQD_TBL_FIELDS.INST_ID]
    return self.displayedIds

  def activateRegionCopier(self, selectedIds: OneDArr=None):
    if selectedIds is None:
      selectedIds = self.selectedIds
    if len(selectedIds) == 0: return
    vertsList = self._compMgr.compDf.loc[selectedIds, REQD_TBL_FIELDS.VERTICES]
    self.regionCopier.resetBaseData(vertsList, selectedIds)
    self.regionCopier.active = True

  def finishRegionCopier(self, keepResult=True):
    if not keepResult: return
    newComps = self._compMgr.compDf.loc[self.regionCopier.regionIds].copy()
    regionOffset = self.regionCopier.offset.astype(int)
    # TODO: Truncate vertices that lie outside image boundaries
    # Invalid if any verts are outside image bounds
    truncatedCompIds = []
    # imShape_xy = self._mainImgArea.image.shape[:2][::-1]
    for idx in newComps.index:
      newVerts = []
      for verts in newComps.at[idx, REQD_TBL_FIELDS.VERTICES]:
        verts = verts + regionOffset
        # goodVerts = np.all(verts >= imShape_xy, 1)
        # if not np.all(goodVerts):
        #   verts = verts[goodVerts,:]
        #   truncatedCompIds.append(idx)
        newVerts.append(verts)
      newComps.at[idx, REQD_TBL_FIELDS.VERTICES] = ComplexXYVertices(newVerts)
    # truncatedCompIds = np.unique(truncatedCompIds)
    if self.regionCopier.inCopyMode:
      self._compMgr.addComps(newComps)
      self.activateRegionCopier(self.regionCopier.regionIds)
    else: # Move mode
      self.regionCopier.erase()
      self._compMgr.addComps(newComps, PRJ_ENUMS.COMP_ADD_AS_MERGE)
    # if len(truncatedCompIds) > 0:
    #   warn(f'Some regions extended beyond image dimensions. Boundaries for the following'
    #        f' components were altered: {truncatedCompIds}', UserWarning)

  def exportCompOverlay(self, outFile='', toClipboard=False):
    """
    :param outFile:
      pType: filepicker
      existing: False
    """
    pm = self._mainImgArea.imgItem.getPixmap()
    painter = QtGui.QPainter(pm)
    self.regionPlot.paint(painter)
    if outFile:
      # if outFile.endswith('svg'):
      #   svgr = QtSvg.QSvgRenderer(outFile)
      #   svgr.render(painter)
      #   painter.end()
      # else:
      painter.end()
      pm.save(outFile)
    if toClipboard:
      QtWidgets.QApplication.clipboard().setImage(pm.toImage())
    return pm


  def resetCompBounds(self):
    self.regionPlot.resetRegionList()


from collections import deque, namedtuple
from warnings import warn

import numpy as np
import pandas as pd
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore
from utilitys import PrjParam

from s3a import PRJ_SINGLETON, PRJ_CONSTS as CNST, XYVertices, REQD_TBL_FIELDS as RTF, \
  ComplexXYVertices
from s3a.models.s3abase import S3ABase
from s3a.processing.algorithms import _historyMaskHolder
from s3a.structures import BlackWhiteImg
from s3a.views.regions import MultiRegionPlot, makeMultiRegionDf
from .base import TableFieldPlugin
from ..constants import PRJ_ENUMS
from ..generalutils import getCroppedImg, showMaskDiff
from ..graphicsutils import RegionHistoryViewer


class _REG_ACCEPTED: pass
buffEntry = namedtuple('buffentry', 'id_ vertices')

class VerticesPlugin(TableFieldPlugin):
  name = 'Vertices'

  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()
    cls.procCollection = PRJ_SINGLETON.imgProcClctn.createProcessorForClass(cls, cls.name + ' Processor')

    cls.dock.addEditors([cls.procCollection])

  def __init__(self):
    super().__init__()
    self.region = MultiRegionPlot()
    self.region.hide()
    self.firstRun = True
    self.playbackWindow = RegionHistoryViewer()
    self.regionBuffer = deque(maxlen=CNST.PROP_UNDO_BUF_SZ.value)

  def attachWinRef(self, win: S3ABase):
    win.mainImg.addItem(self.region)

    def resetRegBuff(_, newSize):
      newBuff = deque(maxlen=newSize)
      newBuff.extend(self.regionBuffer)
      self.regionBuffer = newBuff
    mainBufSize = PRJ_SINGLETON.generalProps.params.child(win.__groupingName__, 'maxLength')
    mainBufSize.sigValueChanged.connect(resetRegBuff)

    def fill():
      """Completely fill the focused region mask"""
      if self.mainImg.compSer is None: return
      filledImg = np.ones(self.mainImg.image.shape[:2], dtype='uint16')
      self.updateRegionFromMask(filledImg)
    def clear():
      """
      Clear the vertices in the focused image
      """
      self.updateRegionFromDf(None)
    def clearProcessorHistory():
      """
      Each time an update is made in the processor, it is saved so algorithmscan take
      past edits into account when performing their operations. Clearing that history
      will erase algorithm knowledge of past edits.
      """
      _historyMaskHolder[0].fill(0)

    funcLst = [self.resetFocusedRegion, fill, clear, clearProcessorHistory]
    paramLst = [CNST.TOOL_RESET_FOC_REGION, CNST.TOOL_FILL_FOC_REGION,
                CNST.TOOL_CLEAR_FOC_REGION, CNST.TOOL_CLEAR_HISTORY]
    for func, param in zip(funcLst, paramLst):
      self.registerFunc(func, btnOpts=param)

    win.mainImg.registerDrawAction([CNST.DRAW_ACT_ADD, CNST.DRAW_ACT_REM], self._run_drawAct)
    win.mainImg.addTools(self.toolsEditor)
    self.vb: pg.ViewBox = win.mainImg.getViewBox()
    super().attachWinRef(win)

  def updateFocusedComp(self, newComp:pd.Series = None):
    if self.mainImg.compSer[RTF.INST_ID] == -1:
      self.updateRegionFromDf(None)
      return
    oldId = self.mainImg.compSer[RTF.INST_ID]
    self.updateRegionFromDf(self.mainImg.compSer_asFrame)
    self.clearFocusedPen_Fill()
    if newComp is None or oldId != newComp[RTF.INST_ID]:
      self.firstRun = True

  def clearFocusedPen_Fill(self):
    plt: MultiRegionPlot = self.win.compDisplay.regionPlot
    ids = plt.regionData[PRJ_ENUMS.FIELD_FOCUSED].to_numpy()
    for name, fn in zip(['pen', 'brush'], [pg.mkPen, pg.mkBrush]):
      clrs = plt.data[name]
      clrs[ids] = fn('0000')
      plt.data[name] = clrs
    plt.updateSpots(plt.data)

  def _run_drawAct(self, verts: XYVertices, param: PrjParam):
    # noinspection PyTypeChecker
    verts : XYVertices = verts.astype(int)
    if param == CNST.DRAW_ACT_ADD:
      self.run(fgVerts=verts)
    else:
      self.run(bgVerts=verts)

  def run(self, fgVerts: XYVertices=None, bgVerts: XYVertices=None):
    vertsDict = {}
    if fgVerts is not None:
      vertsDict['fgVerts'] = fgVerts
    if bgVerts is not None:
      vertsDict['bgVerts'] = bgVerts
    img = self.mainImg.image
    if img is None:
      compGrayscale = None
      compMask = None
    else:
      compGrayscale = self.region.toGrayImg(img.shape[:2])
      compMask = compGrayscale > 0
    # TODO: When multiple classes can be represented within focused image, this is where
    #  change will have to occur
    # Clip viewrange to min view axis area instead of max, which will happen internally
    # otherwise
    viewbox = self.mainImg.viewboxSquare()
    newGrayscale = self.curProcessor.run(
      image=img,
      prevCompMask=compMask,
      **vertsDict,
      firstRun=self.firstRun,
      viewbox=XYVertices(viewbox),
      prevCompVerts=self.mainImg.compSer[RTF.VERTICES]
    ).astype('uint8')

    self.firstRun = False
    if not np.array_equal(newGrayscale, compGrayscale):
      self.updateRegionFromMask(newGrayscale)

  @PRJ_SINGLETON.actionStack.undoable('Modify Focused Component')
  def updateRegionFromDf(self, newData: pd.DataFrame=None, offset: XYVertices=None):
    """
    Updates the current focused region using the new provided vertices
    :param newData: Dataframe to use.If *None*, the image will be totally reset and the component
      will be removed. Otherwise, the provided value will be used. For column information,
      see `makeMultiRegionDf`
    :param offset: Offset of newVerts relative to main image coordinates
    """
    fImg = self.mainImg
    oldSer = fImg.compSer
    if newData is None or np.all(newData[RTF.VERTICES].apply(ComplexXYVertices.isEmpty)):
      newData = makeMultiRegionDf(0)
    if fImg.image is None:
      self.region.clear()
      self.regionBuffer.append(buffEntry(oldSer[RTF.INST_ID], ComplexXYVertices()))
      return
    oldData = self.region.regionData

    if offset is None:
      offset = XYVertices([[0,0]])
    # 0-center new vertices relative to FocusedImage image
    # Make a copy of each list first so we aren't modifying the
    # original data
    centeredData = newData.copy()
    if np.any(offset != 0):
      centeredVerts = []
      for complexVerts in centeredData[RTF.VERTICES]:
        newVertList = ComplexXYVertices()
        for vertList in complexVerts:
          newVertList.append(vertList+offset)
        centeredVerts.append(newVertList)
      centeredData[RTF.VERTICES] = centeredVerts

    if oldData.equals(centeredData):
      return
    else:
      self.region.resetRegionList(newRegionDf=centeredData)
      self.region.focusById(centeredData.index)
      buffVerts = ComplexXYVertices()
      for inner in centeredData[RTF.VERTICES]: buffVerts.extend(inner)
      self.regionBuffer.append(buffEntry(oldSer[RTF.INST_ID], buffVerts))
      yield
    if (fImg.compSer.loc[RTF.INST_ID] != oldSer.loc[RTF.INST_ID]
        or fImg.image is None):
      self.win.changeFocusedComp(oldSer)
    self.region.resetRegionList(oldData)
    self.regionBuffer.pop()

  def updateRegionFromMask(self, mask: BlackWhiteImg, offset=None):
    if offset is None:
      offset = XYVertices([0,0])
    df = makeMultiRegionDf(vertices=[ComplexXYVertices.fromBwMask(mask)])
    self.updateRegionFromDf(df, offset=offset)

  def acceptChanges(self, overrideVerts: ComplexXYVertices=None):
    # Add in offset from main image to VertexRegion vertices
    newVerts = overrideVerts or self.collapseRegionVerts()
    ser = self.mainImg.compSer
    ser[RTF.VERTICES] = newVerts
    self.updateFocusedComp()

  def collapseRegionVerts(self, simplify=True):
    """
    Region can consist of multiple separate complex vertices. However, the focused series
    can only contain one list of complex vertices. This function collapses all data in self.region
    into one list of complex vertices.

    :param simplify: Overlapping regions can be simplified by converting to and back from
      an image. This can be computationally intensive at times, in which case `simplify` can
      be set to *False*
    """
    outVerts = ComplexXYVertices([verts for cplxVerts in self.region.regionData[RTF.VERTICES] for verts in cplxVerts])
    if simplify:
      outVerts = ComplexXYVertices.fromBwMask(outVerts.toMask())
    return outVerts

  def clearFocusedRegion(self):
    # Reset drawn comp vertices to nothing
    # Only perform action if image currently exists
    if self.mainImg.compSer is None:
      return
    self.updateRegionFromDf(None)

  def resetFocusedRegion(self):
    """Reset the focused image by restoring the region mask to the last saved state"""
    if self.mainImg.compSer is None:
      return
    self.updateRegionFromDf(self.mainImg.compSer_asFrame)

  def _onActivate(self):
    self.region.show()
    self.clearFocusedPen_Fill()

  def _onDeactivate(self):
    self.region.hide()
    self.win.compDisplay.regionPlot.updateColors()

  def getRegionHistory(self):
    outImgs = []
    bufferRegions = []
    if not self.regionBuffer:
      return None, []
    firstId = self.regionBuffer[-1].id_
    bufferRegions = [buf.vertices for buf in self.regionBuffer if buf.id_ == firstId]

    if not bufferRegions:
      return None, []

    # First find offset and img size so we don't
    # have to keep copying a full image sized output every time
    allVerts = np.vstack([v.stack() for v in bufferRegions])
    initialImg, slices = getCroppedImg(self.mainImg.image, allVerts)
    imShape = initialImg.shape[:2]
    offset = slices[0]
    img = np.zeros(imShape, bool)
    outImgs.append(img)
    for singleRegionVerts in bufferRegions:
      # Copy to avoid screwing up undo buffer!
      copied = ComplexXYVertices([subV - offset for subV in singleRegionVerts])
      img = copied.toMask(imShape, warnIfTooSmall=False)
      outImgs.append(img)
    return initialImg, outImgs


  def playbackRegionHistory(self):
    initialImg, history = self.getRegionHistory()
    if initialImg is None:
      warn('No edits found, nothing to do', UserWarning)
      return
    # Add current state as final result
    history += [history[-1]]
    diffs = [showMaskDiff(o, n) for (o, n) in zip(history, history[1:])]
    self.playbackWindow.setDiffs(diffs)
    self.playbackWindow.displayPlt.setImage(initialImg)
    self.playbackWindow.show()
    self.playbackWindow.raise_()